#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

import sys
import os
import json
import logging

app_dir = os.getcwd()

def get_app_dir():
    return app_dir

_log_levels = {"debug": logging.DEBUG,
               "info": logging.INFO,
               "warning": logging.WARNING,
               "error": logging.ERROR,
               "critical": logging.CRITICAL}
    
##################
# DEFAULT CONFIG #
##################
LOG_FILENAME = os.path.splitext(os.path.split(sys.argv[0])[1])[0] + ".log"
LOG_PATH = get_app_dir()
LOG_LEVEL = "info"
LOG_QUIET = False


###################
# INTERNAL CONFIG #
###################
_SETTINGS_PATH = None
_LOG_FILE_PATH = None
_LOG_LEVEL = logging.INFO
_INITIALIZED = False


def load_config(path=None, override_settings=dict()):
    if not path:
        # try to automatically load a config file

        # start with CWD
        try:
            load_config(path="sapphire.conf", override_settings=override_settings)

        except IOError:
            
            # then try app directory
            try:
                load_config(os.path.join(get_app_dir(), "sapphire.conf"))

            except IOError:
                pass

    else:
        f = open(path, 'r')

        global _SETTINGS_PATH
        _SETTINGS_PATH = path

        try:
            config = json.loads(f.read())
            
            mod_dict = globals()
            for k, v in config.iteritems():
                mod_dict[k] = v

        except ValueError:
            raise

        f.close()

    # apply override settings
    mod_dict = globals()

    for k, v in override_settings.iteritems():
        mod_dict[k] = v


def init_logging():
    # set up logging
    dt_format = '%Y-%m-%dT%H:%M:%S'

    global _LOG_FILE_PATH
    _LOG_FILE_PATH = os.path.join(LOG_PATH, LOG_FILENAME)

    global _LOG_LEVEL
    _LOG_LEVEL = _log_levels[LOG_LEVEL]



    # set up file handler
    logging.basicConfig(filename=_LOG_FILE_PATH, 
                        format='%(asctime)s >>> %(levelname)s %(message)s', 
                        datefmt=dt_format, 
                        level=_LOG_LEVEL)

    global LOG_QUIET
    if not LOG_QUIET:
        # add a console handler to print to the console
        console = logging.StreamHandler()
        console.setLevel(_LOG_LEVEL)
        formatter = logging.Formatter('%(levelname)s %(asctime)s %(message)s', datefmt=dt_format)
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)


def init(override_settings=dict()):
    global _INITIALIZED
    
    if _INITIALIZED:
        return

    try:
        load_config(override_settings=override_settings)
        init_logging()

        global _SETTINGS_PATH

        if _SETTINGS_PATH:
            logging.info("Loaded config from: %s" % (_SETTINGS_PATH))

        else:
            logging.info("Loaded default config")

        logging.info("Log level: '%s' to file: %s" % (LOG_LEVEL, _LOG_FILE_PATH))

    except ValueError as e:
        init_logging()

        logging.error("Parse error in config file: %s" % (e))

    _INITIALIZED = True







#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#
"""Sapphire Data
"""

from elysianfields import *
from sapphire.fields import get_field_for_type


class FileInfoField(StructField):
    def __init__(self, **kwargs):
        fields = [Int32Field(_name="filesize"),
                  StringField(_name="filename", _length=32),
                  Uint8Field(_name="flags"),
                  ArrayField(_name="reserved", _field=Uint8Field, _length=15)]

        super(FileInfoField, self).__init__(_name="file_info", _fields=fields, **kwargs)

class FileInfoArray(ArrayField):
    def __init__(self, **kwargs):
        field = FileInfoField

        super(FileInfoArray, self).__init__(_field=field, **kwargs)

class HandleInfoField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint16Field(_name="handle_size"),
                  Uint8Field(_name="handle_type")]

        super(HandleInfoField, self).__init__(_name="mem_info", _fields=fields, **kwargs)

class HandleInfoArray(ArrayField):
    def __init__(self, **kwargs):
        field = HandleInfoField

        super(HandleInfoArray, self).__init__(_field=field, **kwargs)

class FirmwareInfoField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint32Field(_name="firmware_length"),
                  UuidField(_name="firmware_id"),
                  StringField(_name="os_name", _length=128),
                  StringField(_name="os_version", _length=16),
                  StringField(_name="firmware_name", _length=128),
                  StringField(_name="firmware_version", _length=16)]

        super(FirmwareInfoField, self).__init__(_name="firmware_info", _fields=fields, **kwargs)

class SerialFrameHeader(StructField):
    def __init__(self, **kwargs):
        fields = [Uint16Field(_name="len"),
                  Uint16Field(_name="inverted_len")]

        super(SerialFrameHeader, self).__init__(_name="frame_header", _fields=fields, **kwargs)

class DnsCacheField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="status"),
                  Ipv4Field(_name="ip"),
                  Uint32Field(_name="ttl"),
                  StringField(_name="query")]

        super(DnsCacheField, self).__init__(_name="dns_cache", _fields=fields, **kwargs)

class DnsCacheArray(ArrayField):
    def __init__(self, **kwargs):
        field = DnsCacheField

        super(DnsCacheArray, self).__init__(_field=field, **kwargs)

class ThreadInfoField(StructField):
    def __init__(self, **kwargs):
        fields = [StringField(_name="name", _length=64),
                  Uint16Field(_name="flags"),
                  Uint32Field(_name="addr"),
                  Uint16Field(_name="data_size"),
                  Uint32Field(_name="run_time"),
                  Uint32Field(_name="runs"),
                  Uint16Field(_name="line"),
                  Uint64Field(_name="alarm"),
                  ArrayField(_name="reserved", _field=Uint8Field, _length=24)]

        super(ThreadInfoField, self).__init__(_fields=fields, **kwargs)

class ThreadInfoArray(ArrayField):
    def __init__(self, **kwargs):
        field = ThreadInfoField

        super(ThreadInfoArray, self).__init__(_field=field, **kwargs)

class NTPTimestampField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint32Field(_name="seconds"),
                  Uint32Field(_name="fraction")]

        super(NTPTimestampField, self).__init__(_fields=fields, **kwargs)

    def copy(self):
        return NTPTimestampField()

class KVLinkField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="flags"),
                  Int32Field(_name="source_hash"),
                  Int32Field(_name="dest_hash"),
                  ArrayField(_name="query", _field=Int32Field, _length=8)]

        super(KVLinkField, self).__init__(_fields=fields, **kwargs)

class KVLinkArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVLinkField

        super(KVLinkArray, self).__init__(_field=field, **kwargs)

class KVSendField(StructField):
    def __init__(self, **kwargs):
        fields = [Ipv4Field(_name="ip"),
                  Uint16Field(_name="port"),
                  Int32Field(_name="source_hash"),
                  Int32Field(_name="dest_hash"),
                  Int8Field(_name="ttl")]

        super(KVSendField, self).__init__(_fields=fields, **kwargs)

class KVSendArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVSendField

        super(KVSendArray, self).__init__(_field=field, **kwargs)

class KVReceiveCacheField(StructField):
    def __init__(self, **kwargs):
        fields = [Ipv4Field(_name="ip"),
                  Uint16Field(_name="port"),
                  Int32Field(_name="dest_hash"),
                  Int32Field(_name="data"),
                  Uint16Field(_name="sequence"),
                  Int8Field(_name="ttl")]

        super(KVReceiveCacheField, self).__init__(_fields=fields, **kwargs)

class KVReceiveCacheArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVReceiveCacheField

        super(KVReceiveCacheArray, self).__init__(_field=field, **kwargs)








class KVMetaField(StructField):
    def __init__(self, **kwargs):
        fields = [Int8Field(_name="type"),
                  Uint8Field(_name="array_count"),
                  Uint8Field(_name="flags"),
                  Uint16Field(_name="__var_ptr"),
                  Uint16Field(_name="__notifier_ptr"),
                  StringField(_name="param_name", _length=32)]

        super(KVMetaField, self).__init__(_fields=fields, **kwargs)

class KVMetaArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVMetaField

        super(KVMetaArray, self).__init__(_field=field, **kwargs)


class KVLegacyMetaField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="group"),
                  Uint8Field(_name="id"),
                  Int8Field(_name="type"),
                  Uint8Field(_name="array_count"),
                  Uint8Field(_name="flags"),
                  Uint16Field(_name="__var_ptr"),
                  Uint16Field(_name="__notifier_ptr"),
                  StringField(_name="param_name", _length=32)]

        super(KVLegacyMetaField, self).__init__(_fields=fields, **kwargs)

class KVLegacyMetaArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVLegacyMetaField

        super(KVLegacyMetaArray, self).__init__(_field=field, **kwargs)

class KVParamField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="group"),
                  Uint8Field(_name="id"),
                  Int8Field(_name="type")]

        # look up type
        if 'type' in kwargs:
            valuefield = get_field_for_type(kwargs['type'], _name='param_value')
            fields.append(valuefield)

            if 'param_value' in kwargs:
                valuefield._value = kwargs['param_value']
                kwargs['param_value'] = valuefield._value

        super(KVParamField, self).__init__(_fields=fields, **kwargs)

    def unpack(self, buffer):
        super(KVParamField, self).unpack(buffer)

        buffer = buffer[self.size():]

        # get value field based on type
        try:
            valuefield = get_field_for_type(self.type, _name='param_value')
            valuefield.unpack(buffer)

            self._fields[valuefield._name] = valuefield

        except KeyError:
            print "ERROR:", self.type
            print self
            raise

        return self

class KVParamArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVParamField

        super(KVParamArray, self).__init__(_field=field, **kwargs)

class KVStatusField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="group"),
                  Uint8Field(_name="id"),
                  Int8Field(_name="status")]

        super(KVStatusField, self).__init__(_fields=fields, **kwargs)

class KVStatusArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVStatusField

        super(KVStatusArray, self).__init__(_field=field, **kwargs)

class KVRequestField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="group"),
                  Uint8Field(_name="id"),
                  Int8Field(_name="type")]

        super(KVRequestField, self).__init__(_fields=fields, **kwargs)

        self.__type = None

    # return size of returned param in response to this request
    def paramSize(self):
        if self.__type != self.type:
            self.__type = self.type
            self.__param_size = KVParamField(type=self.type).size()

        return self.__param_size

    # return size of returned status in response to this request
    def statusSize(self):
        return KVStatusField().size()

class KVRequestArray(ArrayField):
    def __init__(self, **kwargs):
        field = KVRequestField

        super(KVRequestArray, self).__init__(_field=field, **kwargs)

    # return size of returned param in response to this request
    def paramSize(self):
        return sum([request.paramSize() for request in self._fields])

    # return size of returned status in response to this request
    def statusSize(self):
        return sum([request.statusSize() for request in self._fields])


raw_events = """


#define EVENT_ID_NONE                           0
#define EVENT_ID_LOG_INIT                       1
#define EVENT_ID_LOG_RECORD                     2

#define EVENT_ID_SYS_ASSERT                     10
#define EVENT_ID_SIGNAL                         11
#define EVENT_ID_WATCHDOG_KICK                  12

#define EVENT_ID_DEBUG_0                        20
#define EVENT_ID_DEBUG_1                        21
#define EVENT_ID_DEBUG_2                        22
#define EVENT_ID_DEBUG_3                        23
#define EVENT_ID_DEBUG_4                        24
#define EVENT_ID_DEBUG_5                        25
#define EVENT_ID_DEBUG_6                        26
#define EVENT_ID_DEBUG_7                        27

#define EVENT_ID_CPU_SLEEP                      200
#define EVENT_ID_CPU_WAKE                       201

#define EVENT_ID_FFS_GARBAGE_COLLECT            400
#define EVENT_ID_FFS_WEAR_LEVEL                 401

#define EVENT_ID_TIMER_ALARM_MISS               500

#define EVENT_ID_MEM_DEFRAG                     600

#define EVENT_ID_CMD_START                      700
#define EVENT_ID_CMD_TIMEOUT                    701
#define EVENT_ID_CMD_CRC_ERROR                  702

#define EVENT_ID_THREAD_ID                      800
#define EVENT_ID_FUNC_ENTER                     801
#define EVENT_ID_FUNC_EXIT                      802
#define EVENT_ID_STACK_POINTER                  803

"""

def parse_raw_events(events):
    event_dict = {}
    lines = events.split('\n')

    for line in lines:
        tokens = line.split()

        if len(tokens) < 3:
            continue

        event_dict[int(tokens[2])] = tokens[1].lower().replace('event_id_', '')

    return event_dict

event_lookup = parse_raw_events(raw_events)


class EventField(StructField):
    def __init__(self, **kwargs):
        fields = [Uint16Field(_name="event_id"),
                  Uint16Field(_name="param"),
                  Uint32Field(_name="timestamp")]

        super(EventField, self).__init__(_fields=fields, **kwargs)

    def get_event_str(self):
        return event_lookup[self.event_id]

    event_str = property(get_event_str)


class EventArray(ArrayField):
    def __init__(self, **kwargs):
        field = EventField

        super(EventArray, self).__init__(_field=field, **kwargs)



if __name__ == '__main__':
    pass



#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

from elysianfields import *


SAPPHIRE_TYPE_NONE             = 0
SAPPHIRE_TYPE_BOOL             = 1
SAPPHIRE_TYPE_UINT8            = 2
SAPPHIRE_TYPE_INT8             = 3
SAPPHIRE_TYPE_UINT16           = 4
SAPPHIRE_TYPE_INT16            = 5
SAPPHIRE_TYPE_UINT32           = 6
SAPPHIRE_TYPE_INT32            = 7
SAPPHIRE_TYPE_UINT64           = 8
SAPPHIRE_TYPE_INT64            = 9
SAPPHIRE_TYPE_FLOAT            = 10

SAPPHIRE_TYPE_STRING128        = 40
SAPPHIRE_TYPE_MAC48            = 41
SAPPHIRE_TYPE_MAC64            = 42
SAPPHIRE_TYPE_KEY128           = 43
SAPPHIRE_TYPE_IPv4             = 44
SAPPHIRE_TYPE_STRING512        = 45
SAPPHIRE_TYPE_STRING32         = 46
SAPPHIRE_TYPE_STRING64         = 47

SAPPHIRE_TYPE_MISMATCH         = -6

type_registry = {
    SAPPHIRE_TYPE_NONE: None,
    SAPPHIRE_TYPE_BOOL: BooleanField,
    SAPPHIRE_TYPE_UINT8: Uint8Field,
    SAPPHIRE_TYPE_INT8: Int8Field,
    SAPPHIRE_TYPE_UINT16: Uint16Field,
    SAPPHIRE_TYPE_INT16: Int16Field,
    SAPPHIRE_TYPE_UINT32: Uint32Field,
    SAPPHIRE_TYPE_INT32: Int32Field,
    SAPPHIRE_TYPE_UINT64: Uint64Field,
    SAPPHIRE_TYPE_INT64: Int64Field,
    SAPPHIRE_TYPE_FLOAT: FloatField,

    SAPPHIRE_TYPE_STRING128: String128Field,
    SAPPHIRE_TYPE_STRING32: String32Field,
    SAPPHIRE_TYPE_STRING64: String64Field,
    SAPPHIRE_TYPE_STRING512: String512Field,
    SAPPHIRE_TYPE_MAC48: Mac48Field,
    SAPPHIRE_TYPE_MAC64: Mac64Field,
    SAPPHIRE_TYPE_KEY128: Key128Field,
    SAPPHIRE_TYPE_IPv4: Ipv4Field,
    SAPPHIRE_TYPE_MISMATCH: ErrorField,
}

type_id_registry = {
    'none': SAPPHIRE_TYPE_NONE,
    'bool': SAPPHIRE_TYPE_BOOL,
    'uint8': SAPPHIRE_TYPE_UINT8,
    'int8': SAPPHIRE_TYPE_INT8,
    'uint16': SAPPHIRE_TYPE_UINT16,
    'int16': SAPPHIRE_TYPE_INT16,
    'uint32': SAPPHIRE_TYPE_UINT32,
    'int32': SAPPHIRE_TYPE_INT32,
    'uint64': SAPPHIRE_TYPE_UINT64,
    'int64': SAPPHIRE_TYPE_INT64,
    'float': SAPPHIRE_TYPE_FLOAT,

    'string128': SAPPHIRE_TYPE_STRING128,
    'string32': SAPPHIRE_TYPE_STRING32,
    'string64': SAPPHIRE_TYPE_STRING64,
    'string512': SAPPHIRE_TYPE_STRING512,
    'mac48': SAPPHIRE_TYPE_MAC48,
    'mac64': SAPPHIRE_TYPE_MAC64,
    'key128': SAPPHIRE_TYPE_KEY128,
    'ipv4': SAPPHIRE_TYPE_IPv4,
}

def get_type_id(type_name):
    return type_id_registry[type_name]

def get_field_for_type(t, **kwargs):
    try:
        return type_registry[t](**kwargs)

    except KeyError:
        type_id = type_id_registry[t]
        return type_registry[type_id](**kwargs)




from elysianfields import *


class Payload(StructField):

    msg_type_format = None
    msg_type = 0
    fields = []

    def __init__(self, **kwargs):
        fields_copy = deepcopy(self.fields)

        super(Payload, self).__init__(_fields=fields_copy, **kwargs)

        if self.msg_type_format:
            self._local_msg_type_format = deepcopy(self.msg_type_format)
            self._local_msg_type_format._value = self.msg_type

    def size(self):
        field_size = super(Payload, self).size()

        if self.msg_type_format:
            field_size += self._local_msg_type_format.size()

        return field_size

    def unpack(self, buffer):
        if self.msg_type_format:
            # slice past message type if present
            buffer = buffer[self._local_msg_type_format.size():]

        super(Payload, self).unpack(buffer)

        return self

    def pack(self):
        s = ""

        if self.msg_type_format:
            s = self._local_msg_type_format.pack()

        s += super(Payload, self).pack()

        return s


class Protocol(object):

    class NullPayload(Payload):
        msg_type = 0
        fields = []

    msg_type_format = Uint8Field()

    def __init__(self):
        self.messages = self.get_msgs()
        self.__msg_dict = {}

        self._local_msg_type_format = deepcopy(self.msg_type_format)

        for message in self.messages:
            message.msg_type_format = self._local_msg_type_format
            self.__msg_dict[message.msg_type] = message

    def get_msgs(self):
        return [m[1] for m in inspect.getmembers(self)
                    if hasattr(m[1], '__bases__') and Payload in m[1].__bases__]

    def unpack(self, data):
        # get message type from data
        msg_type = self._local_msg_type_format.unpack(data)._value

        # initialize and unpack message
        try:
            msg = self.__msg_dict[msg_type]().unpack(data)

        except KeyError:
            raise

        return msg




# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

import util
from ribbon import Ribbon

import json
import logging
import signal
import sys
import os
import time
import Queue


def load_settings(filename='settings.json'):
    settings = {}

    try:
        with open(filename, 'r') as f:
            settings = json.loads(f.read())

    except IOError:
        pass

    return settings


class App(object):
    def __init__(self,
                 name=None,
                 enable_logging=True,
                 max_log_entries=256,
                 host=None):

        self.settings = load_settings()

        if enable_logging:
            log_filename = None

            if name:
                log_filename = '%s.log' % (name)

                if 'LOG_HOME' in self.settings:
                    log_filename = os.path.join(self.settings['LOG_HOME'], log_filename)

            util.setup_basic_logging(filename=log_filename)


        signal.signal(signal.SIGTERM, self.sigterm_handler)

        try:
            self.version = self.settings['VERSION']

        except KeyError:
            self.version = '0.0.0'

        self.process_id = os.getpid()
        self.hostname = util.hostname()
        self.ip = util.ip()
        self.command = sys.argv[0]

        self.log_entries = []
        self.max_log_entries = max_log_entries

        if name:
            self.name = name

        else:
            self.name = "%d@%s" % (self.process_id, self.hostname)

        logging.info("Starting app: %s" % (self.name))
        logging.info("v: %s" % (self.version))
        logging.info("Process ID: %d" % (self.process_id))

    def to_dict(self):
        d = {
            'name': self.name,
            'command': self.command,
            'process_id': self.process_id,
            'hostname': self.hostname,
            'ip': self.ip,
        }

        return d

    def sigterm_handler(self, signum, frame):
        logging.info("Received SIGTERM")
        sys.exit()

    def stop(self):
        # shutdown all running ribbons
        Ribbon.shutdown()

        logging.info("Shutdown complete")

    def run(self):
        try:
            while True:
                try:
                    pass

                except Exception as e:
                    logging.exception("App: unexpected exception: %s" % (e))

                time.sleep(1.0)

        except KeyboardInterrupt:
            logging.info("Received KeyboardInterrupt")

        except SystemExit:
            pass

        try:
            logging.debug("Shutting down...")

            self.stop()

        except Exception as e:
            logging.exception("App: unexpected exception on shutdown: %s" % (e))


class test(Ribbon):
    def loop(self):
        self.wait(5.0)
        logging.info("stuff")


if __name__ == "__main__":
    app = App()

    r = test()

    app.run()



# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

import fnmatch

MAX_FILENAMES = 10000
MAX_PATTERNS = 10000

cache = {}

# @profile
def fnmatch2(filename, pattern):
    if filename in cache:
        if pattern in cache[filename]:
            return cache[filename][pattern]

        else:
            if len(cache[filename]) > MAX_PATTERNS:
                cache[filename].popitem()

    else:
        if len(cache) > MAX_FILENAMES:
            cache.popitem()

        cache[filename] = {}

    if fnmatch.fnmatch(filename, pattern):
        cache[filename][pattern] = True
        return True

    else:
        cache[filename][pattern] = False
        return False

# @profile
def filter2(names, pattern):
    if pattern == '*':
        return names

    return [n for n in names if fnmatch2(n, pattern)]


import time

# @profile
def test_fnmatch():
    start = time.time()
    for i in xrange(1000000):
        fnmatch.fnmatch('stuff', '*')
        fnmatch.fnmatch('things.stuff', '*.stuff')
        fnmatch.fnmatch('cat.meow', 'cat*.meow')

    elapsed = time.time() - start

    print elapsed * 1000.0


# @profile
def test_fnmatch2():
    start = time.time()
    for i in xrange(1000000):
        fnmatch2('stuff', '*')
        fnmatch2('things.stuff', '*.stuff')
        fnmatch2('cat.meow', 'cat*.meow')

    elapsed = time.time() - start

    print elapsed * 1000.0


def test_mem_usage():
    for i in xrange(1000):
        for j in xrange(100):
            fnmatch2('stuff.%d' % (i), '*%d' % (j))

    for i in xrange(1000):
        for j in xrange(100):
            fnmatch2('stuff.%d' % (i), '*%d' % (j))

# test_fnmatch()
# test_fnmatch2()

# test_mem_usage()

# print "DONE"
# time.sleep(10.0)





#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

import sapphire.udpx.udpx as udpx
import serial
from elysianfields import *

try:
    import serial.tools.list_ports
except ImportError:
    pass

import crcmod
import sapphiredata
import socket
import struct
import time
import threading
from Queue import Queue

MFG_NAME = "Sapphire Open Systems"

# legacy serial protocol:
ATMEL_VID = 0x03eb
ATMEL_PID = 0x2404

SERIAL_SOF = 0xfd
SERIAL_ACK = 0xa1
SERIAL_NAK = 0x1b

# legacy:
CMD_USART_VERSION_LEGACY = 0x01

# UDP serial transport:
CMD_USART_VERSION = 0x02
CMD_USART_UDP_SOF = 0x85
CMD_USART_UDP_ACK = 0x97
CMD_USART_UDP_NAK = 0x14


CHROMATRON_VID = 0xF055
CHROMATRON_PID = 0x73D5

CMD2_UDP_PORT = 16386

class ChannelException(Exception):
    def __init__(self, value = None):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ChannelTimeoutException(ChannelException):
    pass

class ChannelErrorException(ChannelException):
    pass

class ChannelUnreachableException(ChannelException):
    pass

class ChannelInvalidHostException(ChannelException):
    pass

class Channel(object):
    def __init__(self, host, channel_type='none'):
        self.host = host
        self.channel_type = channel_type

    def __del__(self):
        self.close()

    def open(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def read(self):
        raise NotImplementedError

    def write(self, data, port=None):
        raise NotImplementedError

    def settimeout(self, timeout=None):
        raise NotImplementedError

class NullChannel(Channel):
    def __init__(self, host):
        super(NullChannel, self).__init__(host)

    def open(self):
        pass

    def close(self):
        pass

    def read(self):
        pass

    def write(self, data):
        pass

    def settimeout(self, timeout=None):
        pass


class UdpxChannel(Channel):

    def __init__(self, host):
        super(UdpxChannel, self).__init__(host, 'network')

        self.host = host
        self.sock = udpx.ClientSocket()

    def open(self):
        pass

    def close(self):
        self.sock.shutdown()
        self.sock.close()

    def read(self):
        try:
            data, host = self.sock.recvfrom()

            # check host address
            if host[0] != self.host[0]:
                raise ChannelInvalidHostException(self.host)

            else:
                # set host so we have the host port
                self.host = host

        except socket.timeout:
            raise ChannelTimeoutException

        except socket.error:
            raise ChannelUnreachableException

        return data

    def write(self, data, port=None):
        try:
            self.sock.sendto(data, self.host)

        except socket.timeout:
            raise ChannelTimeoutException(self.host)

        except socket.error:
            raise ChannelUnreachableException

    def settimeout(self, timeout):
        self.sock.settimeout(timeout)


class UdpxClientPoolChannel(Channel):

    POOL_SIZE = 4

    __q = Queue(maxsize=POOL_SIZE)

    def __init__(self, host):
        super(UdpxClientPoolChannel, self).__init__(host, 'network')

        self.timeout = None
        self.sock = None

        self.q = self.__q

    def open(self):
        pass

    def close(self):
        if self.sock:
            self.sock.shutdown()

    def getSock(self):
        # block on queue
        self.q.put(None)

        # create socket
        self.sock = udpx.ClientSocket()

        # set timeout
        if self.timeout != None:
            self.sock.settimeout(self.timeout)

    def returnSock(self):
        self.sock = None

        # remove dummy Q item
        self.q.get()

    def read(self):
        data = self.read_data
        self.read_data = None

        return data

    def write(self, data, port=None):
        self.getSock()

        try:
            #print "Sent %4d > %s" % (len(data), self.host)
            self.sock.sendto(data, self.host)

            self.read_data, host = self.sock.recvfrom()
            #print "Recv %4d > %s" % (len(self.read_data), host)

            # check host address
            if host[0] != self.host[0]:
                raise ChannelInvalidHostException(self.host)

            else:
                # set host so we have the host port
                self.host = host

        except socket.timeout:
            raise ChannelTimeoutException(self.host)

        except socket.error:
            raise ChannelUnreachableException

        finally:
            self.returnSock()


    def settimeout(self, timeout=None):
        self.timeout = timeout


UART_RX_BUF_SIZE = 255

class SerialChannel(Channel):
    def __init__(self, host):
        super(SerialChannel, self).__init__(host, 'serial')

        self.open(host)

        # set up crc function
        self.crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')

    def open(self, host):
        self.port = serial.Serial(host, baudrate=2000000)
        self.settimeout(timeout=2.0)
        self.host = self.port.port

    def close(self):
        self.port.close()

    def _read_data(self, length):
        start = time.time()
        data = ''

        while len(data) < length:
            data += self.port.read(length)

            if (len(data) < length) and ((time.time() - start) > self.port.timeout):
                # raise ChannelTimeoutException()
                return data

        return data

    def read(self):
        try:
            tries = 4

            while tries > 0:

                tries -= 1

                header = sapphiredata.SerialFrameHeader()

                # wait for header data
                header_data = self._read_data(header.size())

                # unpack header
                header.unpack(header_data)

                # check header
                if ( ( header.len != ( ~header.inverted_len & 0xffff ) ) ):
                    print "read header error"
                    continue

                # receive data
                data = self._read_data(header.len)
                if len(data) != header.len:
                    print "invalid data length"
                    print header.len, len(data)

                    raise ChannelErrorException()


                # receive CRC
                crc = struct.unpack('>H', self._read_data(2))[0]

                # check crc
                if self.crc_func(data) == crc:
                    return data

                print "read crc error: %x %x" % (crc, self.crc_func(data))
                print header.len, len(data)

                raise ChannelErrorException()

            if tries == 0:
                print "read tries exceeded"
                raise ChannelErrorException()

        except serial.SerialTimeoutException:
            print "read timeout"
            raise ChannelTimeoutException

        except struct.error:
            # print "read struct error"
            raise ChannelErrorException


    def _write_data(self, data):
        while len(data) > 0:
            count = len(data)

            if count > UART_RX_BUF_SIZE:
               count = UART_RX_BUF_SIZE

            self.port.write(chr(count))
            space = ord(self.port.read(1))

            if space < count:
                count = space

            if count == 0:
                time.sleep(0.002)
                continue

            self.port.write(data[:count])

            # slice data
            data = data[count:]

    def write(self, data, port=None):
        tries = 4

        while tries > 0:

            tries -= 1

            try:
                # flush serial input buffer
                self.port.flushInput()

                self._write_data(struct.pack('B', SERIAL_SOF))

                # set up header
                header = sapphiredata.SerialFrameHeader(len=len(data),
                                                        inverted_len=(~len(data) & 0xffff))

                self._write_data(header.pack())

                # send data
                self._write_data(data)

                # compute crc
                crc = self.crc_func(data)

                # send CRC
                self._write_data(struct.pack('>H', crc)) # note the big endian!

                # wait for ACK
                response = struct.unpack('B', self.port.read(1))[0]

                if response == SERIAL_ACK:
                    return

                else:
                    error = struct.unpack('B', self.port.read(1))[0]
                    print "error: %d" % (error)

                    time.sleep(0.5)

            except Exception as e:
                if tries == 0:
                    raise

                # try to reopen port
                self.close()
                self.open(self.host)


        if tries == 0:
            print "write tries exceeded"
            raise ChannelErrorException()

    def settimeout(self, timeout):
        self.port.timeout = timeout


class CmdUsartUDPHeaderLegacy(StructField):
    def __init__(self, **kwargs):
        fields = [Uint8Field(_name="version"),
                  Uint16Field(_name="lport"),
                  Uint16Field(_name="rport"),
                  Uint8Field(_name="reserved"),
                  Uint16Field(_name="data_len"),
                  Uint16Field(_name="header_crc")]

        super(CmdUsartUDPHeaderLegacy, self).__init__(_name="cmd_usart_udp_header_legacy", _fields=fields, **kwargs)

class CmdUsartUDPHeader(StructField):
    def __init__(self, **kwargs):
        fields = [Uint16Field(_name="lport"),
                  Uint16Field(_name="rport"),
                  Uint16Field(_name="data_len"),
                  Uint16Field(_name="header_crc")]

        super(CmdUsartUDPHeader, self).__init__(_name="cmd_usart_udp_header", _fields=fields, **kwargs)


class LegacySerialUDPChannel(Channel):
    def __init__(self, host):
        super(LegacySerialUDPChannel, self).__init__(host, 'serial_udp')

        self.open(host)

        # set up crc function
        self.crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')

        self.lport = CMD2_UDP_PORT
        self.rport = CMD2_UDP_PORT

    def open(self, host):
        self.port = serial.Serial(host, baudrate=2000000)
        self.settimeout(timeout=2.0)
        self.host = self.port.port

    def close(self):
        self.port.close()

    def _read_data(self, length):
        start = time.time()
        data = ''

        while len(data) < length:
            data += self.port.read(length)

            if (len(data) < length) and ((time.time() - start) > self.port.timeout):
                # raise ChannelTimeoutException()
                return data

        return data

    def read(self):
        try:
            header = CmdUsartUDPHeaderLegacy()

            # wait for SOF
            sof = struct.unpack('<B', self._read_data(1))[0]

            if sof != CMD_USART_UDP_SOF:
                print sof
                raise ChannelErrorException()

            # wait for header data
            header_data = self._read_data(header.size())

            # unpack header
            header.unpack(header_data)

            header_crc = header.header_crc
            header.header_crc = 0
            computed_crc = self.crc_func(header.pack())

            # check header
            if header_crc != computed_crc:
                print "read header error"
                raise ChannelErrorException()

            # receive data
            data = self._read_data(header.data_len)
            if len(data) != header.data_len:
                print "invalid data length"
                print header.data_len, len(data)

                raise ChannelErrorException()

            # receive CRC
            crc = struct.unpack('<H', self._read_data(2))[0]

            # check crc
            if self.crc_func(data) == crc:
                return data

            print "read crc error: %x %x" % (crc, self.crc_func(data))
            print header.len, len(data)

            raise ChannelErrorException()

        except serial.SerialTimeoutException:
            print "read timeout"
            raise ChannelTimeoutException

        except struct.error:
            # print "read struct error"
            raise ChannelErrorException


    def _write_data(self, data):
        while len(data) > 0:
            count = len(data)

            if count > UART_RX_BUF_SIZE:
               count = UART_RX_BUF_SIZE

            self.port.write(chr(count))
            space = ord(self.port.read(1))

            if space < count:
                count = space

            if count == 0:
                time.sleep(0.002)
                continue

            self.port.write(data[:count])

            # slice data
            data = data[count:]

    def write(self, data, port=None):
        tries = 4

        if port:
            rport = port
        else:
            rport = self.rport

        while tries > 0:

            tries -= 1

            try:
                # flush serial input buffer
                self.port.flushInput()

                self._write_data(struct.pack('B', CMD_USART_UDP_SOF))

                # set up header
                header = CmdUsartUDPHeaderLegacy(
                            version=CMD_USART_VERSION_LEGACY,
                            lport=self.lport,
                            rport=rport,
                            data_len=len(data))

                crc = self.crc_func(header.pack())
                header.header_crc = crc

                self._write_data(header.pack())

                # send data
                self._write_data(data)

                # compute crc
                crc = self.crc_func(data)

                # send CRC
                self._write_data(struct.pack('<H', crc))

                # wait for ACK
                response = struct.unpack('B', self.port.read(1))[0]

                if response == CMD_USART_UDP_ACK:
                    return

                else:
                    error = struct.unpack('B', self.port.read(1))[0]
                    print "error: %d" % (error)

                    time.sleep(0.5)

            except Exception as e:
                if tries == 0:
                    raise

                # try to reopen port
                self.close()
                self.open(self.host)


        if tries == 0:
            print "write tries exceeded"
            raise ChannelErrorException()

    def settimeout(self, timeout):
        self.port.timeout = timeout


class SerialUDPChannel(Channel):
    def __init__(self, host):
        super(SerialUDPChannel, self).__init__(host, 'serial_udp')

        # set up crc function
        self.crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')

        self.lport = 1
        self.rport = 1

    def open(self, host):
        self.port = serial.Serial(host, baudrate=2000000)
        self.settimeout(timeout=2.0)
        self.host = self.port.port

    def close(self):
        self.port.close()

    def _read_data(self, length):
        start = time.time()
        data = ''

        while len(data) < length:
            data += self.port.read(length)

            if (len(data) < length) and ((time.time() - start) > self.port.timeout):
                # raise ChannelTimeoutException()
                return data

        return data

    def test(self):
        self.write('', tries=1)

    def read(self):
        try:
            header = CmdUsartUDPHeader()

            # wait for SOF
            sof = struct.unpack('<B', self._read_data(1))[0]

            if sof != CMD_USART_UDP_SOF:
                print sof
                raise ChannelErrorException()

            # wait for version
            version = struct.unpack('<B', self._read_data(1))[0]

            if version != CMD_USART_VERSION:
                print version
                raise ChannelErrorException()

            # wait for header data
            header_data = self._read_data(header.size())

            # unpack header
            header.unpack(header_data)

            header_crc = header.header_crc
            header.header_crc = 0
            computed_crc = self.crc_func(header.pack())

            # check header
            if header_crc != computed_crc:
                print "read header error"
                raise ChannelErrorException()

            # receive data
            data = self._read_data(header.data_len)
            if len(data) != header.data_len:
                print "invalid data length"
                print header.data_len, len(data)

                raise ChannelErrorException()

            # receive CRC
            crc = struct.unpack('<H', self._read_data(2))[0]

            # check crc
            if self.crc_func(data) == crc:
                return data

            print "read crc error: %x %x" % (crc, self.crc_func(data))
            print header.len, len(data)

            raise ChannelErrorException()

        except serial.SerialTimeoutException:
            print "read timeout"
            raise ChannelTimeoutException

        except struct.error:
            # print "read struct error"
            raise ChannelErrorException


    def _write_data(self, data):
        self.port.write(data)
        
    def write(self, data, port=None, tries=None):
        if tries == None:
            tries = 4

        if port:
            rport = port
        else:
            rport = self.rport

        while tries > 0:
            tries -= 1

            try:
                # flush serial input buffer
                self.port.flushInput()

                self._write_data(struct.pack('B', CMD_USART_UDP_SOF))

                self._write_data(struct.pack('B', CMD_USART_VERSION))

                # set up header
                header = CmdUsartUDPHeader(
                            lport=self.lport,
                            rport=rport,
                            data_len=len(data))

                crc = self.crc_func(header.pack())
                header.header_crc = crc

                self._write_data(header.pack())

                # check if sending data.
                # we can send an empty message as a comms check.
                if len(data) > 0:
                    # send data
                    self._write_data(data)

                    # compute crc
                    crc = self.crc_func(data)

                    # send CRC
                    self._write_data(struct.pack('<H', crc))

                # wait for ACK
                response = struct.unpack('B', self.port.read(1))[0]

                if response == CMD_USART_UDP_ACK:
                    return

                else:
                    error = struct.unpack('B', self.port.read(1))[0]
                    print "error: %d" % (error)

                    time.sleep(0.5)

            except Exception as e:
                # if tries == 0:
                    # raise

                # try to reopen port
                self.close()
                self.open(self.host)


        if tries == 0:
            print "write tries exceeded"
            raise ChannelErrorException()

    def settimeout(self, timeout):
        self.port.timeout = timeout




class USBChannel(SerialChannel):
    def __init__(self, host):
        super(USBChannel, self).__init__(host)

        self.open(host)

    def close(self):
        try:
            super(USBChannel, self).close()

        except AttributeError:
            pass

        except TypeError:
            # we'll get a TypeError when __del__ is called
            pass

    def open(self, host):
        try:
            self.port = serial.Serial(host, baudrate=2000000)
            self.settimeout(timeout=2.0)
            self.host = self.port.port

        except serial.serialutil.SerialException:
            raise ChannelInvalidHostException("Could not find attached device")


class LegacyUSBUDPChannel(LegacySerialUDPChannel):
    def __init__(self, host):
        super(LegacyUSBUDPChannel, self).__init__(host)

        self.open(host)

    def close(self):
        try:
            super(LegacyUSBUDPChannel, self).close()

        except AttributeError:
            pass

        except TypeError:
            # we'll get a TypeError when __del__ is called
            pass

    def open(self, host):
        try:
            self.port = serial.Serial(host, baudrate=2000000)
            self.settimeout(timeout=5.0)
            self.host = self.port.port

        except serial.serialutil.SerialException:
            raise ChannelInvalidHostException("Could not find attached device")


class USBUDPChannel(SerialUDPChannel):
    def __init__(self, host):
        super(USBUDPChannel, self).__init__(host)

        self.open(host)

    def close(self):
        try:
            super(USBUDPChannel, self).close()

        except AttributeError:
            pass

        except TypeError:
            # we'll get a TypeError when __del__ is called
            pass

    def open(self, host):
        try:
            self.port = serial.Serial(host, baudrate=2000000)
            self.settimeout(timeout=0.5)
            self.host = self.port.port

        except serial.serialutil.SerialException:
            raise
            # raise ChannelInvalidHostException("Could not find attached device: %s" % (host))



class UDPSerialBridge(threading.Thread):
    def __init__(self, channel, rport):
        super(UDPSerialBridge, self).__init__()

        self.channel = channel
        self.rport = rport
        self.daemon = True

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(('localhost', 0))

        self.port = self.sock.getsockname()[1]

        super(UDPSerialBridge, self).start()

    def run(self):
        while True:
            try:
                self.sock.settimeout(1.0)
                data, host = self.sock.recvfrom(4096)

                # print "SOCK RECV", len(data)

                self.channel.write(data, port=self.rport)

                # print "SERIAL WRITE", len(data)

                try:
                    response = self.channel.read()

                    # print "SERIAL READ", len(response)

                    self.sock.sendto(response, host)

                    # print "SOCK SEND", len(response)

                except Exception as e:
                    # flush
                    while True:
                        try:
                            self.sock.settimeout(0.1)
                            data, host = self.sock.recvfrom(4096)
                            
                        except socket.timeout:
                            break
    
        
            except socket.timeout:
                pass

            except Exception as e:
                pass
            

def createChannel(host, port=None):
    if host == 'usb':
        comport = None
        porttype = None

        # scan ports
        ports = list(serial.tools.list_ports.comports())

        for port in ports:
            # if port.vid == ATMEL_VID and port.pid == ATMEL_PID and port.manufacturer == MFG_NAME:
            #     comport = port.device
            #     porttype = 'legacy'
            #     break

            # elif port.vid == CHROMATRON_VID and port.pid == CHROMATRON_PID and port.manufacturer == MFG_NAME:
            #     comport = port.device
            #     porttype = 'udp_transport'
            #     break

            if port.vid == CHROMATRON_VID and port.pid == CHROMATRON_PID:
                    comport = port.device
                    porttype = 'udp_transport'
                    break

        if comport == None:
            raise ChannelInvalidHostException("Could not find attached device")

        if porttype == 'legacy':
            return USBChannel(comport)

        elif porttype == 'udp_transport':
            ch = USBUDPChannel(comport)

            try:
                ch.test()

                return ch

            except ChannelErrorException:
                ch.close()

                # fallback to legacy
                # return LegacyUSBUDPChannel(comport)

    try:
        socket.inet_aton(host)
        # return UdpxClientPoolChannel(host=(host, port))
        return UdpxChannel(host=(host, port))

    except:
        return SerialChannel(host)


if __name__ == '__main__':
    ports = list(serial.tools.list_ports.comports())

    for port in ports:
        print port.device, port.vid, port.pid, port.manufacturer, port.product

    # chan = createChannel('usb')
    # print chan

    # chan.write('meow')

    # print chan.read()






#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

import os
import sys
import shutil
import logging
import json
import multiprocessing
import subprocess
import shlex
import uuid
import crcmod
import struct
import argparse
import hashlib
import urllib
import tarfile
import zipfile
from datetime import datetime
import hashlib
import zipfile
import ConfigParser
from pprint import pprint
import firmware_package

import settings

from intelhex import IntelHex

import global_settings

from sapphire.devices.sapphiredata import KVMetaField
from sapphire.common import util
from catbus import catbus_string_hash
from fnvhash import fnv1a_32


class SettingsParseException(Exception):
    pass

class ProjectNotFoundException(Exception):
    pass

class AppZipNotFound(Exception):
    pass


KV_PREFIX = '__KV__'
lowercase = list('abcdefghijklmnopqrstuvwxyz')
uppercase = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
digits = list('1234567890')
KV_CHARS = lowercase
KV_CHARS.extend(uppercase)
KV_CHARS.extend(['_'])
KV_CHARS.extend(digits)
# KV_HEADER = '_kv_hashes.h'
    
SETTINGS_DIR = os.path.dirname(os.path.abspath(global_settings.__file__))
SETTINGS_FILE = 'settings.json'
PROJECTS_FILE_NAME = 'projects.json'
PROJECTS_FILE_PATH = firmware_package.data_dir()
PROJECTS_FILE = os.path.join(PROJECTS_FILE_PATH, PROJECTS_FILE_NAME)
BUILD_CONFIGS_DIR = os.path.join(os.getcwd(), 'configurations')
BUILD_CONFIGS_FWID_FILE = os.path.join(BUILD_CONFIGS_DIR, 'fwid.json')
BASE_DIR = os.getcwd()
MASTER_HASH_DB_FILE = os.path.join(PROJECTS_FILE_PATH, 'kv_hashes.json')
TOOLS_DIR = os.path.join(PROJECTS_FILE_PATH, 'tools')

def get_build_package_dir():
    return firmware_package.firmware_package_dir('build')


tools_linux = [
    ('http://downloads.arduino.cc/tools/', 'avrdude-6.0.1-arduino5-x86_64-pc-linux-gnu.tar.bz2'),
    ('http://downloads.arduino.cc/tools/', 'avr-gcc-4.9.2-atmel3.5.3-arduino2-x86_64-pc-linux-gnu.tar.bz2')
]

tools_darwin = [
    ('http://downloads.arduino.cc/tools/', 'avrdude-6.0.1-arduino5-i386-apple-darwin11.tar.bz2'),
    ('http://downloads.arduino.cc/tools/', 'avr-gcc-4.9.2-atmel3.5.3-arduino2-i386-apple-darwin11.tar.bz2')
]

tools_windows = [
    ('http://downloads.arduino.cc/tools/', 'avrdude-6.0.1-arduino5-i686-mingw32.zip'),
    ('http://downloads.arduino.cc/tools/', 'avr-gcc-4.9.2-atmel3.5.3-arduino2-i686-mingw32.zip')
]

build_tool_archives = {
    'linux': tools_linux,
    'linux2': tools_linux,
    'darwin': tools_darwin,
    'win32': tools_windows
}

def get_tools():
    try:
        os.makedirs(TOOLS_DIR)

    except OSError:
        pass # already exists

    cwd = os.getcwd()

    os.chdir(TOOLS_DIR)

    for tool in build_tool_archives[sys.platform]:
        urllib.URLopener().retrieve(tool[0] + tool[1], tool[1])

        if tool[1].endswith('.tar.bz2'):
            tar = tarfile.open(tool[1])
            tar.extractall()
            tar.close()

        elif tool[1].endswith('.zip'):
            zf = zipfile.ZipFile(tool[1])
            zf.extractall()
            zf.close()

        else:
            raise Exception("no decompressor available")

    os.chdir(cwd)

def runcmd(cmd, tofile=False, tolog=True):
    logging.debug(cmd)

    try:
        p = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    except OSError as e:
        logging.error("Error running command: %s" % (cmd))
        raise

    output = ''

    for line in p.stdout.readlines():
        output += line

    if output != '':
        if not tofile and tolog:
            logging.warn("\n" + output)

        elif tofile:
           f = open(tofile, 'w')
           f.write(output)
           f.close()

    p.wait()

    return output


def get_builder(target_dir, target_type, build_loader=False, fnv_hash=True):
    builder = Builder(target_dir, target_type, fnv_hash=fnv_hash)

    modes = {"os": OSBuilder, "loader": LoaderBuilder, "app": AppBuilder, "lib": LibBuilder, "exe": ExeBuilder}

    return modes[builder.settings["BUILD_TYPE"]](target_dir, target_type, build_loader=build_loader, fnv_hash=fnv_hash)

class Builder(object):
    def __init__(self, target_dir, target_type=None, build_loader=False, fnv_hash=True):
        self.target_dir = target_dir
        self.target_type = target_type

        self.build_loader = build_loader

        try:
            self.settings = self.get_settings()

        except AttributeError:
            self.settings = {}

        try:
            self.proj_name = str(self.settings["PROJ_NAME"]) # convert from unicode

        except KeyError:
            self.proj_name = None

        try:
            self.fwid = self.settings["FWID"]

        except KeyError:
            self.fwid = None

        if fnv_hash:
            self.settings["HASH"] = "fnv"

        else:
            self.settings["HASH"] = "kv"            

        try:
            self.libraries = self.settings["LIBRARIES"]
        except KeyError:
            self.libraries = list()

        try:
            self.includes = self.settings["INCLUDES"]
        except KeyError:
            self.includes = list()

        try:
            self.defines = self.settings["DEFINES"]
        except KeyError:
            self.defines = list()

        if self.build_loader:
            self.defines.append("BOOTLOADER")

        try:
            self.source_dirs = self.settings["EXTRA_SOURCE_DIRS"]
        except KeyError:
            self.source_dirs = list()

        # check if recursive build
        try:
            if bool(self.settings["RECURSIVE_BUILD"]):
                for root, dirs, files in os.walk(self.target_dir):
                    self.source_dirs.append(root)

        except KeyError:
            pass

        # add target dir
        self.source_dirs.append(self.target_dir)

    def get_tools(self):
        tools = build_tool_archives[sys.platform]

    def init_logging(self):
        try:
            os.remove(self.settings["LOG_FILENAME"])

        except:
            pass

        file_logger = logging.FileHandler(filename=self.settings["LOG_FILENAME"])
        file_logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s >>> %(levelname)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
        file_logger.setFormatter(formatter)
        logging.getLogger('').addHandler(file_logger)

    def get_settings(self):
        app_settings = self._get_local_settings()

        # check if app settings override target type
        if "TARGET" in app_settings:
            self.target_type = app_settings["TARGET"]

        self.settings_dir = os.path.join(BASE_DIR, 'targets', self.target_type)

        settings = self._get_default_settings()

        # override defaults with app settings
        for k, v in app_settings.iteritems():
            if k == "LIBRARIES":
                settings[k].extend(v)

            else:
                settings[k] = v

        if "CC" not in settings:
            settings["CC"] = os.path.join(TOOLS_DIR, 'avr', 'bin', 'avr-gcc')

        if "BINTOOLS" not in settings:
            settings["BINTOOLS"] = os.path.join(TOOLS_DIR, 'avr', 'bin')

        if 'FULL_NAME' not in settings:
            try:
                settings['FULL_NAME'] = settings['PROJ_NAME']

            except KeyError:
                settings['FULL_NAME'] = 'unknown'

        return settings

    def _get_default_settings(self):
        try:
            return json.loads(open(os.path.join(self.settings_dir, SETTINGS_FILE)).read())

        except ValueError:
            raise SettingsParseException

    def _get_local_settings(self):
        # load app settings file
        try:
            f = open(os.path.join(self.target_dir, SETTINGS_FILE))

        # file does not exist
        except IOError:
            return dict()

        try:
            # parse settings
            app_settings = json.loads(f.read())

        except ValueError:
            raise SettingsParseException

        return app_settings

    def list_source(self):
        source_files = []

        for d in self.source_dirs:
            for f in os.listdir(d):
                if f.endswith('.c'):
                    fpath = os.path.join(d, f)
                    # prevent duplicates
                    if fpath not in source_files:
                        source_files.append(fpath)

        return source_files

    def list_headers(self):
        source_files = []

        for d in self.source_dirs:
            for f in os.listdir(d):
                if f.endswith('.h'):
                    fpath = os.path.join(d, f)
                    # prevent duplicates
                    if fpath not in source_files:
                        source_files.append(fpath)

        return source_files

    def create_source_hash(self):
        source_files = self.list_source()
        hashed_files = {}

        for f in source_files:
            m = hashlib.sha256()
            m.update(open(f).read())
            hashed_files[f] = m.hexdigest()

        return hashed_files

    def update_source_hash_file(self):
        hash_file = open("source_hash.json", 'w+')

        hash_file.write(json.dumps(self.create_source_hash()))
        hash_file.close()

    def get_source_hash_file(self):
        try:
            hash_file = open("source_hash.json")

        except IOError:
            return {}

        return json.loads(hash_file.read())

    def get_buildnumber(self):
        try:
            f = open(os.path.join(self.target_dir, self.settings["BUILD_NUMBER_FILE"]), 'r')
            build_number = f.read()
            f.close()

        except IOError:
            f = open(os.path.join(self.target_dir, self.settings["BUILD_NUMBER_FILE"]), 'w')
            build_number = "0"
            f.write(build_number)
            f.close()

        return int(build_number)

    def set_buildnumber(self, value):
        f = open(os.path.join(self.target_dir, self.settings["BUILD_NUMBER_FILE"]), 'w')
        f.write("%d" % (value))
        f.close()

    buildnumber = property(get_buildnumber, set_buildnumber)

    def get_version(self):
        s = "%s.%04d" % (str(self.settings["PROJ_VERSION"]), self.buildnumber)
        return s

    version = property(get_version)

    def scan_file_for_kv(self, filename):
        with open(filename, 'r') as f:
            data = f.read()

        hashes = {}

        # search for prefix
        index = data.find(KV_PREFIX)

        while index >= 0:

            # scan to end
            terminal_index = index + len(KV_PREFIX)

            while terminal_index < len(data):
                if data[terminal_index] not in KV_CHARS:
                    break

                terminal_index += 1

            kv_string = data[index + len(KV_PREFIX):terminal_index]
            define_string = KV_PREFIX + kv_string

            hashes[define_string] = catbus_string_hash(kv_string)

            index = data.find(KV_PREFIX, index + 1)

        return hashes

    def create_kv_hashes(self):
        files = self.list_source()
        files.extend(self.list_headers())

        # process source in all included projects as well
        for include in self.includes:
            b = get_project_builder(include, target=self.target_type)

            files.extend(b.list_source())
            files.extend(b.list_headers())

        hashes = {}

        for f in files:
            hashes.update(self.scan_file_for_kv(f))

        # with open(os.path.join(self.target_dir, KV_HEADER), 'w+') as f:
        #     f.write('/*THIS FILE IS AUTOMATICALLY GENERATED!*/\n')
        #     f.write('/*DO NOT EDIT THIS FILE!*/\n')
        #     f.write('#ifndef _KV_HASHES_H_\n')
        #     f.write('#define _KV_HASHES_H_\n')
        #     f.write('#include <inttypes.h>\n')
        #     f.write('#include "catbus_common.h"\n')

        #     for k, v in hashes.iteritems():
        #         f.write('#define %s ((catbus_hash_t32)%s)\n' % (k, v))
    
        #     f.write('#endif\n')
        
        # update hashes in master file
        db_hashes = {}

        try:
            with open(MASTER_HASH_DB_FILE, 'r') as f:
                try:
                    db_hashes.update(json.loads(f.read()))

                except ValueError:
                    pass

        except IOError:
            pass

        db_hashes.update(hashes)

        with open(MASTER_HASH_DB_FILE, 'w+') as f:
            f.write(json.dumps(db_hashes, indent=4, separators=(',', ': '))) 


        return hashes

    def clean(self):
        logging.info('Cleaning %s' % (self.target_dir))

        try:
            os.remove(os.path.join(self.target_dir, 'source_hash.json'))

        except OSError:
            pass

        try:
            os.remove(os.path.join(self.target_dir, 'manifest.txt'))

        except OSError:
            pass

        for f in os.listdir(self.target_dir):
            for filetype in self.settings["CLEAN_FILES"]:
                if f.lower().endswith(filetype):
                    try:
                        os.remove(os.path.join(self.target_dir, f))

                        logging.info('> Removing %s' % (f))

                    except:
                        raise

            if f in self.settings["CLEAN_DIRS"]:
                try:
                    shutil.rmtree(os.path.join(self.target_dir, f), True)
                    logging.info('> Removing %s' % (f))

                except:
                    raise

    def build(self):
        logging.info("Building %s for target: %s" % (self.settings["PROJ_NAME"], self.target_type))

        self.pre_process()
        self.compile()
        self.link()
        self.post_process()

        save_project_info(self.proj_name, self.target_dir)

    def pre_process(self):
        # inc build number
        self.buildnumber += 1
            
        # get KV hashes and add to defines
        hashes = self.create_kv_hashes()

        for k, v in hashes.iteritems():
            self.defines.append('%s=((catbus_hash_t32)%s)' % (k, v))

        # self.settings["C_FLAGS"].append('-fdollars-in-identifiers')

    def get_libraries(self, current_libs={}):
        libs = current_libs

        for lib in self.libraries:
            if lib in current_libs:
                continue

            libs[lib] = get_project_builder(lib, target=self.target_type)

            libs.update(libs[lib].get_libraries(current_libs=libs))

        return libs

    def build_libs(self):
        libs = self.get_libraries()

        for lib, builder in libs.iteritems():
            logging.info("Building library %s" % (lib))
        
            builder.clean()
            builder.build()

    def compile(self):
        logging.info("Compiling %s" % (self.proj_name))

        # save working dir
        cwd = os.getcwd()

        # change to target dir
        os.chdir(self.target_dir)

        try:
            os.mkdir(self.settings["OBJ_DIR"])

        except OSError:
            pass

        try:
            os.mkdir(self.settings["DEP_DIR"])

        except OSError:
            pass

        multicore = True

        if multicore:
            pool = multiprocessing.Pool()

        current_hash = self.create_source_hash()
        last_hash = self.get_source_hash_file()

        source_files = self.list_source()

        filtered_source_files = [f for f in source_files if f not in last_hash or last_hash[f] != current_hash[f]]

        for source_file in filtered_source_files:

            # get dir and filename components
            source_path, source_fname = os.path.split(source_file)

            # this is a bit of a hack to get somewhat sane source file
            # names in the __FILE__ macro, but still work with
            # the recursive build option.
            compile_path = source_file

            if source_path == os.getcwd():
                compile_path = source_fname

            logging.info("Compiling file %s" % (compile_path))

            # build command string
            cmd = '"' + self.settings["CC"] + '"'
            cmd += ' -c %s ' % (compile_path)

            for include in self.includes:
                include_dir = get_project_builder(include, target=self.target_type).target_dir

                cmd += '-I%s ' % (include_dir)

            for include in self.libraries:
                include_dir = get_project_builder(include, target=self.target_type).target_dir

                cmd += '-I%s ' % (include_dir)
                cmd += '-D%s ' % (include.upper())

            # add extra source dirs to include
            for source_dir in self.source_dirs:
                cmd += '-I%s ' % (source_dir)

            for define in self.defines:
                cmd += '-D%s ' % (define)

            for flag in self.settings["C_FLAGS"]:
                cmd += flag + ' '

            cmd += '%(DEP_DIR)/%(SOURCE_FNAME).o.d' + ' '
            cmd += '-o ' + '%(OBJ_DIR)/%(SOURCE_FNAME).o' + ' '

            cmd = cmd.replace('%(OBJ_DIR)', self.settings["OBJ_DIR"])
            cmd = cmd.replace('%(SOURCE_FNAME)', source_fname)
            cmd = cmd.replace('%(DEP_DIR)', self.settings["DEP_DIR"])

            # replace windows path separators with unix
            cmd = cmd.replace('\\', '/')

            if multicore:
                pool.apply_async(runcmd, args = (cmd, ))

            else:
                runcmd(cmd)

        if multicore:
            pool.close()
            pool.join()

        # update hash
        self.update_source_hash_file()

        # change back to working dir
        os.chdir(cwd)

    def link(self):
        pass

    def post_process(self):
        pass


class ConfigBuilder(Builder):
    def __init__(self, *args, **kwargs):
        self.build_config = kwargs['build_config']
        del kwargs['build_config']

        super(ConfigBuilder, self).__init__(os.getcwd(), **kwargs)

        self.proj_name = self.build_config['PROJ_NAME']
        self.fwid = self.build_config['FWID']

        self.app_builder = get_project_builder(self.build_config['BASE_PROJ'], target=self.target_type)
            
        # change app's FWID to the config FWID
        self.app_builder.settings['FWID'] = self.fwid
        
        self.app_builder.settings['PROJ_NAME'] = self.build_config['PROJ_NAME']
        self.app_builder.settings['PROJ_VERSION'] = self.build_config['PROJ_VERSION']
        self.app_builder.settings['FULL_NAME'] = self.build_config['FULL_NAME']


        self.lib_builders = []
        for lib in self.build_config['LIBS']:
            builder = get_project_builder(lib, target=self.target_type)

            self.lib_builders.append(builder)


    def build(self):
        logging.info("Building configuration %s for target: %s" % (self.proj_name, self.target_type))

        includes = []
        inits = []

        # build libraries
        for builder in self.lib_builders:

            builder.clean()
            builder.build()
            
            includes.extend(builder.settings["LIB_INCLUDES"])
            inits.append(builder.settings["LIB_INIT"])
            
            self.app_builder.libraries.append(builder.proj_name)


        lib_init_filename = os.path.join(self.app_builder.target_dir, '__libs.c')

        lib_init = """
#include "sapphire.h"
"""
    
        for include in includes:
            lib_init += '\n#include "%s"\n' % (include)

        lib_init += """

void libs_v_init( void ){

"""

        for init in inits:
            if len(init) > 0:
                lib_init += '\n%s();\n' % (init)

        lib_init += '\n}\n'


        # create lib init file
        with open(lib_init_filename, 'w+') as f:
            f.write(lib_init)

        # self.app_builder.build_libs()
        self.app_builder.clean()
        try:
            self.app_builder.build()

        except AppZipNotFound: # not a problem in config-driven builds
            pass

        # make sure we have the firmware package dir
        try:
            os.makedirs(get_build_package_dir())

        except OSError:
            pass

        # copy firmware zip
        shutil.copy(os.path.join(self.app_builder.target_dir, '%s.zip' % (self.proj_name)), get_build_package_dir())

        # update build date
        with open(os.path.join(get_build_package_dir(), firmware_package.PUBLISHED_AT_FILENAME), 'w') as f:
            f.write(util.now().isoformat())


class LibBuilder(Builder):
    def __init__(self, *args, **kwargs):
        super(LibBuilder, self).__init__(*args, **kwargs)

        try:
            self.includes.append(self.settings["OS_PROJECT"])
        except KeyError:
            pass

    # def link(self):
    #     logging.info("Linking %s" % (self.proj_name))
    #
    #     # save working dir
    #     cwd = os.getcwd()
    #
    #     # change to target dir
    #     os.chdir(self.target_dir)
    #
    #     # build command string
    #     cmd = self.settings["AR"] + ' '
    #
    #     for flag in self.settings["AR_FLAGS"]:
    #         cmd += flag + ' '
    #
    #     cmd += self.proj_name + '.a' + ' '
    #
    #     obj_dir = self.settings["OBJ_DIR"]
    #
    #     for source in self.list_source():
    #         source_path, source_fname = os.path.split(source)
    #         cmd += obj_dir + '/' + source_fname + '.o' + ' '
    #
    #     # replace windows path separators with unix
    #     cmd = cmd.replace('\\', '/')
    #
    #     runcmd(cmd)
    #
    #     # change back to working dir
    #     os.chdir(cwd)


class OSBuilder(LibBuilder):
    def __init__(self, *args, **kwargs):
        super(OSBuilder, self).__init__(*args, **kwargs)

class HexBuilder(Builder):
    def __init__(self, *args, **kwargs):
        super(HexBuilder, self).__init__(*args, **kwargs)

    def link(self):
        logging.info("Linking %s" % (self.proj_name))

        # build command string
        # enclose in quotes so we can handle spaces in the command filepath
        cmd = '"' + self.settings["CC"] + '"'
        cmd += ' '

        for flag in self.settings["C_FLAGS"]:
            cmd += flag + ' '

        cmd += '%(DEP_DIR)/%(SOURCE_FNAME).o.d' + ' '

        # save working dir
        cwd = os.getcwd()

        # change to target dir
        os.chdir(self.target_dir)

        obj_dir = self.settings["OBJ_DIR"]

        # source object files
        for source in self.list_source():
            source_path, source_fname = os.path.split(source)
            cmd += obj_dir + '/' + source_fname + '.o' + ' '

        # linker flags
        for flag in self.settings["LINK_FLAGS"]:
            cmd += flag + ' '

        # included libraries
        for lib in self.libraries:
            lib_project = get_project_builder(lib, target=self.target_type)
            # source object files
            for source in lib_project.list_source():
                source_path, source_fname = os.path.split(source)
                cmd += lib_project.target_dir + '/' + obj_dir + '/' + source_fname + '.o' + ' '

            # lib_dir = get_project_builder(lib, target=self.target_type).target_dir
            #
            # cmd += '%s ' % (lib_dir + '/' + lib + '.a')

        # include all libraries again, so GCC can handle cross dependencies
        # for lib in self.libraries:
        #     lib_dir = get_project_builder(lib, target=self.target_type).target_dir
        #
        #     cmd += '%s ' % (lib_dir + '/' + lib + '.a')

        cmd = cmd.replace('%(OBJ_DIR)', obj_dir)
        cmd = cmd.replace('%(DEP_DIR)', self.settings["DEP_DIR"])
        # cmd = cmd.replace('%(SOURCE_FNAME)', self.proj_name)
        cmd = cmd.replace("%(LINKER_SCRIPT)", os.path.join(self.settings_dir, "linker.x"))
        cmd = cmd.replace("%(APP_NAME)", self.settings["PROJ_NAME"])
        cmd = cmd.replace("%(TARGET_DIR)", self.target_dir)

        # replace windows path separators with unix
        cmd = cmd.replace('\\', '/')

        runcmd(cmd)

        logging.info("Generating output files")

        # enclose in quotes so we can handle spaces in the command filepath
        bintools = '"' + self.settings["BINTOOLS"] + '"'
        runcmd(os.path.join(bintools, 'avr-objcopy -O ihex -R .eeprom main.elf main.hex'))
        runcmd(os.path.join(bintools, 'avr-size -B main.elf'))
        runcmd(os.path.join(bintools, 'avr-objdump -h -S -l main.elf'), tofile='main.lss')
        runcmd(os.path.join(bintools, 'avr-nm -n main.elf'), tofile='main.sym')

        # change back to working dir
        os.chdir(cwd)


class AppBuilder(HexBuilder):
    def __init__(self, *args, **kwargs):
        super(AppBuilder, self).__init__(*args, **kwargs)

        # insert OS first
        try:
            self.libraries.insert(0, self.settings["OS_PROJECT"])
        except KeyError:
            pass

    def merge_hex(self, hex1, hex2, target):
        ih = IntelHex(hex1)
        ih.merge(IntelHex(hex2))

        ih.write_hex_file(target)

    def post_process(self):
        logging.info("Processing binaries")

        fw_info_fmt = '<I16s128s16s128s16s'

        fw_info_addr = int(self.settings["FW_INFO_ADDR"], 16)

        # save original dir
        cwd = os.getcwd()

        # change to target dir
        os.chdir(self.target_dir)

        ih = IntelHex('main.hex')

        fwid = uuid.UUID('{' + self.settings["FWID"] + '}')

        # get KV meta start
        kv_meta_addr = fw_info_addr + struct.calcsize(fw_info_fmt)
        kv_meta_len = KVMetaField().size()

        bindata = ih.tobinstr()

        kv_meta_data = []

        while True:
            kv_meta = KVMetaField().unpack(bindata[kv_meta_addr:kv_meta_addr + kv_meta_len])

            kv_meta_addr += kv_meta_len

            if kv_meta.param_name == "kvstart":
                continue

            elif kv_meta.param_name == "kvend":
                break

            else:
                kv_meta_data.append(kv_meta)

        kv_meta_by_hash = {}

        logging.info("Hash type: FNV1A_32")

        # create lookups by 32 bit hash
        index = 0
        for kv in kv_meta_data:
            hash32 = fnv1a_32(str(kv.param_name))

            if hash32 in kv_meta_by_hash:
                raise Exception("Hash collision!")

            kv_meta_by_hash[hash32] = (kv, index)

            index += 1

        # sort indexes
        sorted_hashes = sorted(kv_meta_by_hash.keys())

        # create binary look up table
        kv_index = ''
        for a in sorted_hashes:
            kv_index += struct.pack('<LB', a, kv_meta_by_hash[a][1])
                
        # write to end of hex file
        ih.puts(ih.maxaddr() + 1, kv_index)

        size = ih.maxaddr() - ih.minaddr() + 1

        # get os info
        try:
            os_project = get_project_builder(self.settings["OS_PROJECT"], target=self.target_type)
        except KeyError:
            os_project = ""

        # create firmware info structure
        fw_info = struct.pack(fw_info_fmt,
                                size,
                                fwid.bytes,
                                os_project.proj_name,
                                os_project.version,
                                str(self.settings['PROJ_NAME']),
                                self.version)

        # insert fw info into hex
        ih.puts(fw_info_addr, fw_info)

        # compute crc
        crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')

        crc = crc_func(ih.tobinstr())

        logging.info("size: %d" % (size))
        logging.info("fwid: %s" % (fwid))
        logging.info("fwinfo: %x" % (fw_info_addr))
        logging.info("kv index len: %d" % (len(kv_index)))
        logging.info("crc: 0x%x" % (crc))
        logging.info("os name: %s" % (os_project.proj_name))
        logging.info("os version: %s" % (os_project.version))
        logging.info("app name: %s" % (self.settings['PROJ_NAME']))
        logging.info("app version: %s" % (self.version))

        ih.puts(ih.maxaddr() + 1, struct.pack('>H', crc))

        ih.write_hex_file('main.hex')
        ih.tobinfile('firmware.bin')

        # get loader info
        loader_project = get_project_builder(self.settings["LOADER_PROJECT"], target=self.target_type)

        # create loader image
        loader_hex = os.path.join(loader_project.target_dir, "main.hex")
        self.merge_hex('main.hex', loader_hex, 'loader_image.hex')

        # create sha256 of binary
        sha256 = hashlib.sha256(ih.tobinstr())

        # create manifest file
        data = {
            'name': self.settings['FULL_NAME'],
            'timestamp': datetime.utcnow().isoformat(),
            'sha256': sha256.hexdigest(),
            'fwid': self.settings['FWID'],
            'version': self.version
        }

        with open('manifest.txt', 'w+') as f:
            f.write(json.dumps(data))

        # create firmware zip file
        zf = zipfile.ZipFile('chromatron_main_fw.zip', 'wb')
        zf.write('manifest.txt')
        zf.write('firmware.bin')
        zf.close()

        # create second, project specific zip
        # we'll remove the first zip after
        # we update the firmware tools
        zf = zipfile.ZipFile('%s.zip' % (self.settings['PROJ_NAME']), 'wb')
        zf.write('manifest.txt')
        zf.write('firmware.bin')
        zf.close()


        # change back to original dir
        os.chdir(cwd)

        logging.info("Package dir: %s" % (get_build_package_dir()))

        # make sure we have the firmware package dir
        try:
            os.makedirs(get_build_package_dir())

        except OSError:
            pass


        # copy firmware zip
        try:
            shutil.copy(os.path.join(self.target_dir, '%s.zip' % (self.proj_name)), get_build_package_dir())

        except IOError:
            raise AppZipNotFound

        # update build date
        with open(os.path.join(get_build_package_dir(), firmware_package.PUBLISHED_AT_FILENAME), 'w') as f:
            f.write(util.now().isoformat())


class LoaderBuilder(HexBuilder):
    def __init__(self, *args, **kwargs):
        super(LoaderBuilder, self).__init__(*args, **kwargs)

        # try:
        #     self.libraries.insert(0, self.settings["OS_PROJECT"])
        # except KeyError:
        #     pass

        for lib in self.libraries:
            self.includes.append(lib)

        self.libraries = list()

        try:
            self.includes.append(self.settings["OS_PROJECT"])
        except KeyError:
            pass

        self.settings["LINK_FLAGS"].append("-Wl,--section-start=.text=%s" % (self.settings["LOADER_ADDRESS"]))


class ExeBuilder(Builder):
    def __init__(self, *args, **kwargs):
        super(ExeBuilder, self).__init__(*args, **kwargs)

        # insert OS first
        try:
            self.libraries.insert(0, self.settings["OS_PROJECT"])
        except KeyError:
            pass

    def list_source(self):
        source_files = []

        for lib in self.libraries:
            lib_proj = get_project_builder(lib, target=self.target_type)

            source_files.extend(lib_proj.list_source())

        for d in self.source_dirs:
            for f in os.listdir(d):
                if f.endswith('.c'):
                    fpath = os.path.join(d, f)
                    # prevent duplicates
                    if fpath not in source_files:
                        source_files.append(fpath)

        return source_files

    def link(self):
        logging.info("Linking %s" % (self.proj_name))

        # build command string
        cmd = '"' + self.settings["CC"] + '"'
        cmd += ' '

        for flag in self.settings["C_FLAGS"]:
            cmd += flag + ' '

        cmd += '%(DEP_DIR)/%(SOURCE_FNAME).o.d' + ' '

        # save working dir
        cwd = os.getcwd()

        # change to target dir
        os.chdir(self.target_dir)

        obj_dir = self.settings["OBJ_DIR"]

        source_files = self.list_source()

        # source object files
        for source in source_files:
            source_path, source_fname = os.path.split(source)
            cmd += obj_dir + '/' + source_fname + '.o' + ' '

        # linker flags
        for flag in self.settings["LINK_FLAGS"]:
            cmd += flag + ' '

        # included libraries
        # for lib in self.libraries:
            # lib_proj = get_project_builder(lib, target=self.target_type)

            # lib_dir = get_project_builder(lib, target=self.target_type).target_dir

            # cmd += '%s ' % (lib_dir + '/' + lib + '.a')

        # include all libraries again, so GCC can handle cross dependencies
        # for lib in self.libraries:
            # lib_dir = get_project_builder(lib, target=self.target_type).target_dir

            # cmd += '%s ' % (lib_dir + '/' + lib + '.a')

        cmd = cmd.replace('%(OBJ_DIR)', obj_dir)
        cmd = cmd.replace('%(DEP_DIR)', self.settings["DEP_DIR"])
        # cmd = cmd.replace('%(SOURCE_FNAME)', self.proj_name)
        cmd = cmd.replace("%(LINKER_SCRIPT)", os.path.join(self.settings_dir, "linker.x"))
        cmd = cmd.replace("%(APP_NAME)", self.settings["PROJ_NAME"])
        cmd = cmd.replace("%(TARGET_DIR)", self.target_dir)

        # replace windows path separators with unix
        cmd = cmd.replace('\\', '/')

        runcmd(cmd)

        # change back to working dir
        os.chdir(cwd)


def get_project_info_dir():
    return settings.get_app_dir()

def get_project_info_file():
    return PROJECTS_FILE

def get_build_configs():
    build_configs = {}

    # open/create firmware ID file
    try:
        fwid_f = open(BUILD_CONFIGS_FWID_FILE, 'r')

    except IOError:
        # create file
        fwid_f = open(BUILD_CONFIGS_FWID_FILE, 'w+')
    
    try:
        fwids = json.loads(fwid_f.read())

    except ValueError:
        fwids = {}

    fwid_f.close()


    for filename in os.listdir(BUILD_CONFIGS_DIR):
        filepath = os.path.join(BUILD_CONFIGS_DIR, filename)
        
        configparser = ConfigParser.SafeConfigParser()

        try:
            configparser.read(filepath)

        except ConfigParser.MissingSectionHeaderError:
            continue

        for section in configparser.sections():
            try:
                proj_name       = section
                proj_version    = configparser.get(section, 'proj_version')
                base_proj       = configparser.get(section, 'base_proj')

                try:
                    full_name   = configparser.get(section, 'full_name')

                except ConfigParser.NoOptionError:
                    full_name   = proj_name                

                try:
                    libs        = configparser.get(section, 'libs').split(',')

                except ConfigParser.NoOptionError:
                    libs = []

                # trim whitespace
                libs = [lib.strip() for lib in libs]

                if proj_name in build_configs:
                    raise KeyError("Duplicate project name in build config: %s" % (proj_name))

                # try to get firmware ID
                # create if not found
                if proj_name not in fwids:
                    fwids[proj_name] = str(uuid.uuid4())

                    print "Created FWID %s for project: %s" % (fwids[proj_name], proj_name)
                    
                    # save to file
                    with open(BUILD_CONFIGS_FWID_FILE, 'w+') as f:
                        f.write(json.dumps(fwids))

                build_configs[proj_name] = {'PROJ_NAME': proj_name,
                                            'PROJ_VERSION': proj_version,
                                            'FULL_NAME': full_name,
                                            'FWID': fwids[proj_name],
                                            'BASE_PROJ': base_proj,
                                            'LIBS': libs}

            except ConfigParser.NoOptionError:
                pass


    return build_configs

def get_project_list():
    try:
        f = open(get_project_info_file(), 'r')
        data = f.read()
        f.close()

        return json.loads(data)

    except IOError:
        return dict()

    except ValueError:
        return dict()

def save_project_info(name, target_dir):
    data = get_project_list()
    data[name] = target_dir

    # getdata dir
    data_dir = get_project_info_dir()

    # check if data dir exists
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    f = open(get_project_info_file(), 'w')
    f.truncate(0)
    f.write(json.dumps(data, indent=4, separators=(',', ': ')))
    f.close()

    logging.info("Saving project listing to: %s" % (data_dir))

def remove_project_info(project_name):
    data = get_project_list()

    # check if project exists
    if project_name not in data:
        raise ProjectNotFoundException

    del data[project_name]

    f = open(get_project_info_file(), 'w')
    f.truncate(0)
    f.write(json.dumps(data, indent=4, separators=(',', ': ')))
    f.close()

    logging.info("Removed %s from listing" % (project_name))

def get_project_builder(proj_name=None, fwid=None, target=None, build_loader=False):
    projects = get_project_list()

    try:
        if proj_name:
            return get_builder(projects[proj_name], target, build_loader=build_loader)
            # return Builder(projects[proj_name], target_type=target)

        elif fwid:
            for k, v in projects.iteritems():
                try:
                    builder = get_project_builder(k, target=target)

                except ProjectNotFoundException:
                    pass

                if builder.fwid == fwid:
                    return builder

    except KeyError:
        if proj_name:
            # try to build from config
            configs = get_build_configs()

            if proj_name not in configs:
                raise ProjectNotFoundException(proj_name)

            # build from config
            return ConfigBuilder(build_config=configs[proj_name], target_type=target)

        else:
            raise ProjectNotFoundException(fwid)

    except Exception:
        raise


    raise ProjectNotFoundException

# make a new project in the current working dir
def make_project(proj_name):
    logging.info("Creating new project: %s" % (proj_name))

    # create new firmware ID
    fwid = str(uuid.uuid4())

    # set project dir name
    project_dir = proj_name

    # create project dir
    os.mkdir(project_dir)

    # get path to project template
    template_dir = os.path.join(SETTINGS_DIR, 'project_template')

    # copy project template
    for f in os.listdir(template_dir):
        shutil.copy(os.path.join(template_dir, f), project_dir)

    # open settings file
    with open(os.path.join(project_dir, SETTINGS_FILE), 'r+') as f:
        s = f.read()

        # replace template strings
        s = s.replace('%(PROJ_NAME)', proj_name)
        s = s.replace('%(FWID)', fwid)

        f.truncate(0)
        f.seek(0)
        f.write(s)

    logging.info("New project: '%s' created with firmware ID: %s" % (proj_name, fwid))

    save_project_info(proj_name, os.path.abspath(project_dir))


def main():
    # set global log level
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)

    # add a console handler to print anything INFO and above to the console
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)

    parser = argparse.ArgumentParser(description='SapphireMake')

    parser.add_argument("--clean", "-c", action="store_true", help="clean only")
    parser.add_argument("--project", "-p", action="append", help="projects to build")
    parser.add_argument("--build_all", action="store_true", help="build all config projects")
    parser.add_argument("--target", "-t", action="store", help="set build target")
    parser.add_argument("--dir", "-d", action="append", help="directories to build")
    parser.add_argument("--loader",action="store_true", help="build bootloader")
    parser.add_argument("--list", "-l", action="store_true", help="list projects")
    parser.add_argument("--list_configs", action="store_true", help="list build configurations")
    parser.add_argument("--reset", action="store_true", help="reset project listing")
    parser.add_argument("--discover", action="store_true", help="discover projects")
    parser.add_argument("--create", help="create a new template project")
    parser.add_argument("--unlink", help="unlink a project")
    parser.add_argument("--make_release", action="store_true", help="make current build a release package")
    parser.add_argument("--install_build_tools", action="store_true", help="install build tools")

    args = vars(parser.parse_args())

    logging.info("SapphireMake")

    # set target dir to cwd
    target_dir = os.getcwd()

    logging.info("Target dir: %s" % (target_dir))

    # make sure projects dir exists
    try:
        os.makedirs(PROJECTS_FILE_PATH)

    except OSError:
        pass

    logging.info("Projects dir: %s" % (PROJECTS_FILE_PATH))

    # check if making a release
    if args["make_release"]:
        
        # get current packages
        releases = firmware_package.get_releases()

        release_name = raw_input("Enter release name: ")

        # check if we already have this release
        if release_name in releases:
            print "This release already exists."

            overwrite = raw_input("Enter 'overwrite' to overwrite this release: ")

            if overwrite != 'overwrite':
                print "Release cancelled"
                return

            # erase original release
            shutil.rmtree(firmware_package.firmware_package_dir(release=release_name), True)

        # copy build package to new name
        shutil.copytree(get_build_package_dir(), firmware_package.firmware_package_dir(release=release_name))

        shutil.make_archive(os.path.join(firmware_package.firmware_package_dir(), release_name), 'zip', firmware_package.firmware_package_dir(release=release_name))

        # erase build release
        shutil.rmtree(get_build_package_dir(), True)

        return

    # check if installing build tools
    if args["install_build_tools"]:
        get_tools()
        return

    # check if listing
    if args["list"]:
        projects = get_project_list()

        for key in projects:
            print "%20s %s" % (key, projects[key])

        return

    if args["list_configs"]:
        pprint(get_build_configs())
        return

    # check if reset
    if args["reset"]:
        os.remove(get_project_info_file())
        return

    # check if discovering
    if args["discover"]:
        # load old project listing
        try:
            with open(get_project_info_file(), 'r') as f:
                projects = json.loads(f.read())

        except IOError:
            projects = {}

        for root, dirs, files in os.walk(os.getcwd()):
            if SETTINGS_FILE in files:
                # load settings file                
                with open(os.path.join(root, SETTINGS_FILE)) as f:
                    proj_settings = json.loads(f.read())

                    if 'FWID' not in proj_settings:
                        continue

                    if 'BUILD_TYPE' not in proj_settings:
                        continue

                    try:
                        proj_name = proj_settings['PROJ_NAME']

                        if proj_name == '%(PROJ_NAME)':
                            continue

                        projects[proj_name] = root

                    except KeyError:
                        pass

        with open(PROJECTS_FILE, 'w+') as f:
            f.write(json.dumps(projects))

        print "Found %d projects" % (len(projects))
        print "Saved project file: %s" % (PROJECTS_FILE)


        return

    # check if creating a new project
    if args["create"]:
        make_project(args["create"])
        return

    # check if unlinking a new project
    if args["unlink"]:
        remove_project_info(args["unlink"])
        return

    # check if setting target
    if args["target"]:
        target_type = args["target"]

    else:
        target_type = "chromatron"

    if args["loader"]:
        build_loader = True

    else:
        build_loader = False


    if args["build_all"]:
        configs = get_build_configs()
        
        args["project"] = configs.keys()

    # check if project is given
    if args["project"]:
        for p in args["project"]:
            logging.info("Project: %s" % (p))

            # check if project is in the projects list
            try:
                builder = get_project_builder(p, target=target_type)
            except Exception:
                raise

            # init logging
            builder.init_logging()

            # run clean
            if args["clean"]:
                builder.clean()

            else:
                builder.build()

    # check if directory is given
    elif args["dir"]:
        for d in args["dir"]:
            target_dir = os.path.abspath(d)

            builder = get_builder(target_dir, target_type, build_loader=build_loader)

            # init logging
            builder.init_logging()

            # run clean
            builder.clean()

            if not args["clean"]:
                builder.build()

    # build in current directory
    else:
        builder = get_builder(target_dir, target_type, build_loader=build_loader)

        # init logging
        builder.init_logging()

        # run clean
        builder.clean()

        if not args["clean"]:
            builder.build()



if __name__ == "__main__":
    main()
    


# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

from sapphire.common.util import coerce_string_to_list


def query_dict(d, **kwargs):
    # check for all query (all trumps any other keywords)
    if "_all" in kwargs:
        if kwargs["_all"]:
            return d

        del kwargs["_all"]

    if "_expr" in kwargs:
        if not kwargs["_expr"](d):
            return None

        del kwargs["_expr"]

    contains = False
    if "_contains" in kwargs:
        attr_list = kwargs["_contains"]

        # if passed a single string, make it a list
        attr_list = coerce_string_to_list(attr_list)

        for attr in attr_list:
            # converts keys in dict to string for comparison to work
            if attr not in [str(k) for k in d.keys()]:
                return None

        contains = True
        del kwargs["_contains"]

    # if no other kwargs are passed, we have nothing to match against
    if len(kwargs) == 0:
        if contains:
            return d

        else:
            return None

    # match all kwargs
    for k in kwargs:
        if not k in d or str(d[k]) != str(kwargs[k]):
            return None

    return d



#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

"""Device
"""

from elysianfields import *
from protocols import *
from catbus import get_type_name, Client, CATBUS_DISCOVERY_PORT, catbus_string_hash, NoResponseFromHost, ProtocolErrorException

import sapphiredata
import firmware
import channel

import time
import sys
import datetime
import types
from datetime import datetime, timedelta

from pprint import pprint
import hashlib
import binascii
from UserDict import DictMixin
import json
import base64
import traceback
import crcmod

from sapphire.query import query_dict

import fnmatch

crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')


NTP_EPOCH = datetime(1900, 1, 1)


# do not change these!!!
FILE_TRANSFER_LEN   = 512
MAX_KV_DATA_LEN     = 512

MAX_LATENCIES       = 128

# Key value groups
KV_GROUP_NULL               = 0
KV_GROUP_NULL1              = 254
KV_GROUP_SYS_CFG            = 1
KV_GROUP_SYS_INFO           = 2
KV_GROUP_SYS_STATS          = 3
KV_GROUP_APP_BASE           = 32

KV_GROUP_ALL                = 255

kv_groups = {
    "kv_group_null":        KV_GROUP_NULL,
    "kv_group_null_1":      KV_GROUP_NULL1,
    "kv_group_sys_cfg":     KV_GROUP_SYS_CFG,
    "kv_group_sys_info":    KV_GROUP_SYS_INFO,
    "kv_group_sys_stats":   KV_GROUP_SYS_STATS,
    "kv_group_all":         KV_GROUP_ALL,
}

sys_groups = [KV_GROUP_SYS_CFG, KV_GROUP_SYS_INFO, KV_GROUP_SYS_STATS]


KV_ID_ALL                   = 255

# Key value flags
CATBUS_FLAGS_READ_ONLY      = 0x01
CATBUS_FLAGS_PERSIST        = 0x04
CATBUS_FLAGS_DYNAMIC        = 0x08

# warning flags
SYS_WARN_MEM_FULL           = 0x0001
SYS_WARN_NETMSG_FULL        = 0x0002
SYS_WARN_FLASHFS_FAIL       = 0x0004
SYS_WARN_FLASHFS_HARD_ERROR = 0x0008
SYS_WARN_CONFIG_FULL        = 0x0010
SYS_WARN_CONFIG_WRITE_FAIL  = 0x0020
SYS_WARN_EVENT_LOG_OVERFLOW = 0x0040
SYS_WARN_MISSING_KV_INDEX   = 0x0080
SYS_WARN_SYSTEM_ERROR       = 0x8000

def decode_warnings(flags):
    warnings = list()

    if flags & SYS_WARN_MEM_FULL:
        warnings.append("mem_full")

    if flags & SYS_WARN_NETMSG_FULL:
        warnings.append("netmsg_full")

    if flags & SYS_WARN_FLASHFS_FAIL:
        warnings.append("flashfs_fail")

    if flags & SYS_WARN_FLASHFS_HARD_ERROR:
        warnings.append("flashfs_hard_error")

    if flags & SYS_WARN_CONFIG_FULL:
        warnings.append("config_full")

    if flags & SYS_WARN_CONFIG_WRITE_FAIL:
        warnings.append("config_write_fail")

    if flags & SYS_WARN_EVENT_LOG_OVERFLOW:
        warnings.append("event_log_overflow")

    if flags & SYS_WARN_MISSING_KV_INDEX:
        warnings.append("missing_kv_index")

    if flags & SYS_WARN_SYSTEM_ERROR:
        warnings.append("system_error")

    return warnings

SYS_REBOOT_NORMAL    = 1
SYS_REBOOT_SAFE     = -2
SYS_REBOOT_NOAPP    = -3 # not yet supported
SYS_REBOOT_LOADFW   = -4
SYS_REBOOT_FORMATFS = -5
SYS_REBOOT_RECOVERY = -6



CLI_PREFIX = 'cli_'

DeviceStatus = set(['unknown', 'offline', 'online', 'reboot'])


class DeviceCommsErrorException(Exception):
    pass

class DeviceUnreachableException(DeviceCommsErrorException):
    pass

class DeviceMetaErrorException(Exception):
    pass

class DuplicateKeyNameException(DeviceMetaErrorException):
    pass

class DuplicateKeyIDException(DeviceMetaErrorException):
    pass

class ReadOnlyKeyException(DeviceMetaErrorException):
    pass

class UnrecognizedKeyException(DeviceMetaErrorException):
    pass

class InvalidDataTypeException(DeviceMetaErrorException):
    pass

class InvalidDeviceIDException(DeviceMetaErrorException):
    pass

class KVKey(object):
    def __init__(self,
                 device=None,
                 group=None,
                 id=None,
                 flags=None,
                 type=None,
                 value=None,
                 key=None):

        self._device = device

        self.group = group
        self.id = id
        self.type = type
        self._value = value
        self.key = key

        # decode flags to strings
        self.flags = list()

        if flags & CATBUS_FLAGS_READ_ONLY:
            self.flags.append("read_only")

        if flags & CATBUS_FLAGS_PERSIST:
            self.flags.append("persist")

        if flags & CATBUS_FLAGS_DYNAMIC:
            self.flags.append("dynamic")

    def __str__(self):
        flags = ''

        if "persist" in self.flags:
            flags += 'P'
        else:
            flags += ' '

        if "read_only" in self.flags:
            flags += 'R'
        else:
            flags += ' '

        if "dynamic" in self.flags:
            flags += 'D'
        else:
            flags += ' '

        s = "%32s %6s %8s %s" % (self.key, flags, get_type_name(self.type), str(self._value))
        return s

    def is_sys(self):
        return self.group < KV_GROUP_APP_BASE

    def is_readonly(self):
        return "read_only" in self.flags

    def get_value(self):
        # if value hasn't been initialized, we load from the device
        if self._value == None:
            # internal meta data is auto-updated, so we can toss the return value
            self._device.get_key(self.key)

        return self._value

    def set_value(self, value):
        self._value = value
        self._device.set_key(self.key, self._value)

    value = property(get_value, set_value)


class KVMeta(DictMixin):
    def __init__(self):
        self.kv_items = dict()

    def keys(self):
        return self.kv_items.keys()

    def __getitem__(self, key):
        return self.kv_items[key]

    def __setitem__(self, key, value):
        if key in self.kv_items:
            print "DuplicateKeyNameException: %s" % (key)
            # print key
            # we already have an item here
            # raise DuplicateKeyNameException("DuplicateKeyNameException: %s" % (key))

        self.kv_items[key] = value
        value.key = key

    def __delitem__(self, key):
        del self.kv_items[key]

    def check(self):
        # check for duplicate IDs
        for key in self.kv_items:
            l = len([k for k, v in self.kv_items.iteritems()
                        if v.group == self.kv_items[key].group
                            and v.id == self.kv_items[key].id])

            if l > 1:
                raise DuplicateKeyIDException("DuplicateKeyIDException: %s" % (key))

    def is_system(self, key):
        group = self.kv_items[key].group

        return group in sys_groups


class Device(object):
    def __init__(self,
                 host=None,
                 device_id=None,
                 command_protocol=DeviceCommandProtocol,
                 response_protocol=DeviceCommandResponseProtocol,
                 comm_channel=None):

        super(Device, self).__init__()

        self.host = host
        self.firmware_id = None
        self.firmware_name = ""
        self.os_name = ""
        self.firmware_version = ""
        self.os_version = ""
        self.device_id = device_id
        self.name = "<anon@%s>" % (host)

        self.device_status = 'offline'

        self.object_id = str(self.device_id)

        self._keys = KVMeta()
        self._firmware_info_hash = None

        self._channel = comm_channel

        # if no channel is specified, create one
        if self._channel is None:
            try:
                self._channel = channel.createChannel(self.host, port=DeviceCommandProtocol.PORT)

            except channel.ChannelInvalidHostException:
                raise DeviceUnreachableException

        self.channel_type = self._channel.channel_type
        self._client = None
        self._bridge = None

        if self.channel_type == 'network':
            self._client = Client()
            self._client.connect(self.host, get_meta=False)

        elif self.channel_type == 'serial_udp':
            self._client = Client()
            # set window to 1, so messages will ping pong over the bridge.
            self._client.set_window(1, 1)
            self._bridge = channel.UDPSerialBridge(self._channel, CATBUS_DISCOVERY_PORT)

            try:
                self._client.connect(('localhost', self._bridge.port))

            except Exception as e:
                print "fall back to legacy", 0
                print e
                
                # fail to connect
                self._client = None


        # assign protocols
        self._protocol = command_protocol()
        self._response_protocol = response_protocol()

        # stats
        self.msgs_sent = 0
        self.msgs_recv = 0

        self.latencies = []


    def __str__(self):
        return "Device:%s@%s" % (self.name, self.host)


    # TODO this should move to the console
    def who(self):
        try:
            name = str(self._keys['name'].value)

        except KeyError:
            name = ''

        if len(name) == 0:
            name = "<anon@%s>" % self.host

        s = "%24s IP:%15s FW:%24s Ver:%8s OS:%8s" % \
            (name,
             self.host,
             self.firmware_name,
             self.firmware_version,
             self.os_version)

        return s

    def to_dict(self):
        d = dict()

        for k, v in self._keys.iteritems():
            d[k] = v.value

        d["firmware_name"] = self.firmware_name

        return d

    def query(self, **kwargs):
        return query_dict(self.to_dict(), **kwargs)

    def scan(self, get_all=True):
        self.get_firmware_info()
        try:
            self.get_kv_meta()

            if get_all:
                self.get_all_kv()

            else:
                self.get_key('device_id')
                try:
                    self.get_key('meta_tag_name')

                except KeyError:
                    self.get_key('name')

            if self.device_id == None:
                self.device_id = self._keys["device_id"]._value

            elif self.device_id != self._keys["device_id"]._value:
                raise InvalidDeviceIDException(self._keys["device_id"]._value, self.device_id)

            try:
                if self._keys["meta_tag_name"]._value:
                    self.name = self._keys["meta_tag_name"]._value

            except KeyError:
                if self._keys["name"]._value:
                    self.name = self._keys["name"]._value


        except DuplicateKeyIDException:
            # error is printed by the check function.
            # we pass here since we can get to the file system and fix
            # the firmware without the KV system.
            pass

        return self

    def _send_command(self, cmd, timeout=None):
        try:
            start = time.time()

            if timeout:
                self._channel.settimeout(timeout)

            self._channel.write(cmd.pack())
            
            self.msgs_sent += 1

            data = self._channel.read()

            stop = time.time()

            self.msgs_recv += 1

            latency = stop - start

            self.latencies.append(latency)

            if len(self.latencies) > MAX_LATENCIES:
                self.latencies.pop(0)

            if self.device_status != 'online':
                self.device_status = 'online'

        except channel.ChannelException as e:
            if self.device_status == 'online':
                self.device_status = 'offline'

            # print type(e)
            # print e
            # print cmd
            # traceback.print_exc()

            raise DeviceUnreachableException("Device:%s unreachable" % (self.host))


        response = self._response_protocol.unpack(data)

        if len(data) != response.size():
            raise ValueError

        return response

    def average_latency(self):
        return sum(self.latencies) / float(len(self.latencies))

    def get_cli(self):
        return [f.replace(CLI_PREFIX, '', 1) for f in dir(self)
                if f.startswith(CLI_PREFIX)
                and not f.startswith('_')
                and isinstance(self.__getattribute__(f), types.MethodType)]

    def get_kv_meta(self):
        try:
            kvmeta = self._client.get_meta()

            # reset keys
            self._keys = KVMeta()

            # load keys into meta data
            for key, meta in kvmeta.iteritems():
                self._keys[key] = KVKey(key=key, device=self, group=0, id=meta.hash, flags=meta.flags, type=meta.type)

            # run duplicate ID check
            try:
                self._keys.check()

            except DuplicateKeyIDException as e:
                print e

                raise

        except (NoResponseFromHost, IOError, AttributeError):
            print "fallback to legacy", 1
            self._client = None
        
            data = self.get_file("kvmeta")

            # unpack kv meta data
            kvmeta = sapphiredata.KVLegacyMetaArray().unpack(data)

            # reset keys
            self._keys = KVMeta()

            # load keys into meta data
            for kv in kvmeta:
                self._keys[kv.param_name] = KVKey(key=kv.param_name, device=self, group=kv.group, id=kv.id, flags=kv.flags, type=kv.type)

            # run duplicate ID check
            try:
                self._keys.check()

            except DuplicateKeyIDException as e:
                print e

                raise


    def get_all_kv(self):
        keys = [key for key in self._keys]

        return self.get_kv(*keys)

    def set_key(self, param, value):
        self.set_kv(**{param: value})

    def set_kv(self, **kwargs):
        params = []

        keys = {}
        filter_unchanged = False
        if '_filter_unchanged' in kwargs:
            filter_unchanged = kwargs['_filter_unchanged']
            del kwargs['_filter_unchanged']

        data = {}

        # iterate over keys and create requests for them
        for key in kwargs:
            if filter_unchanged:
                # filter out keys which are not changing
                if kwargs[key] == self._keys[key].value:
                   continue

            # check if key is set to read only
            if 'read_only' in self._keys[key].flags:
                raise ReadOnlyKeyException(key)

            if isinstance(kwargs[key], basestring):
                # we do not support unicode, and things break if leaks into the system
                kwargs[key] = str(kwargs[key])

            if self._client:
                data[key] = kwargs[key]

            else:
                param = sapphiredata.KVParamField(group=self._keys[key].group,
                                                  id=self._keys[key].id,
                                                  type=self._keys[key].type,
                                                  param_value=kwargs[key])
                params.append(param)

                keys[(param.group, param.id)] = key

        if self._client:
            self._client.set_keys(**data)

        else:
            # now we have a list of requests, we need to batch them together such
            # that the params all fit within packet size constraints
            batches = [sapphiredata.KVParamArray() for param in params]

            for batch in batches:
                for param in params:
                    if batch.size() + param.size() < MAX_KV_DATA_LEN:
                        batch.append(param)

                for param in batch:
                    params.remove(param)

            # filter out batches which are empty
            batches = [batch for batch in batches if len(batch) > 0]

            # send each batch
            for batch in batches:
                cmd = self._protocol.SetKV(data=batch.pack())

                #response = self._send_command(cmd)
                response_msg = self._send_command(cmd)
                response = sapphiredata.KVStatusArray().unpack(response_msg.data)

                # parse responses
                for param in response:
                    key = keys[(param.group, param.id)]

                    # check status
                    if param.status >= 0:
                        self._keys[key]._value = kwargs[key]

                        if key == 'name':
                            self.name = kwargs[key]

                    else:
                        raise ValueError

    def get_key(self, param):
        return self.get_kv(param)[param]

    def get_kv(self, *args):
        params = []

        keys = {}
        expanded_keys = []
        responses = {}

        # iterate over keys and create requests for them
        for key in args:
            expanded_keys.extend(fnmatch.filter(self._keys.keys(), key))

        if self._client:
            responses = self._client.get_keys(expanded_keys)
        
            for k, v in responses.iteritems():
                # update internal meta data
                self._keys[k]._value = v

        else:
            for key in expanded_keys:
                param = sapphiredata.KVRequestField(group=self._keys[key].group,
                                                    id=self._keys[key].id,
                                                    type=self._keys[key].type)

                params.append(param)

                keys[(param.group, param.id)] = key

            # now we have a list of requests, we need to batch them together such
            # that the responses all fit within packet size constraints
            batches = [sapphiredata.KVRequestArray() for param in params]

            for batch in batches:
                for param in params:
                    if ( batch.paramSize() + param.paramSize() ) < MAX_KV_DATA_LEN:
                        batch.append(param)

                for param in batch:
                    params.remove(param)

            # filter out batches which are empty
            batches = [batch for batch in batches if len(batch) > 0]

            # request each batch
            for batch in batches:
                cmd = self._protocol.GetKV(data=batch.pack())

                response_msg = self._send_command(cmd)
                try:
                    response = sapphiredata.KVParamArray().unpack(response_msg.data)

                    # parse responses
                    for param in response:
                        key = keys[(param.group, param.id)]

                        responses[key] = param.param_value

                        # update internal meta data
                        self._keys[key]._value = param.param_value

                except KeyError as e:
                    print "KeyError: %s" % (str(e))
                    print "attempting to connect anyway..."

                except struct.error as e:
                    print "struct.error: %s" % (str(e))
                    print "attempting to connect anyway..."

        return responses

    def reset_config(self):
        raise NotImplementedError
        # return self._send_command(self._protocol.ResetCfg())

    def set_security_key(self, key_id, key):
        raise NotImplementedError
        # cmd = self._protocol.SetSecurityKey(key_id=key_id, key=key)

        # return self._send_command(cmd)

    def echo(self, data='\0' * 32):
        if self._client:
            try:
                return self._client.ping()

            except NoResponseFromHost:
                raise DeviceUnreachableException

        else:
            return self._send_command(self._protocol.Echo(echo_data=data))

    def _reboot_cmd(self, cmd):
        try:
            response = self._send_command(cmd)
            self.device_status = "offline"

        except:
            raise

        return response

    def reboot(self):
        if self._client:
            return self.set_key('reboot', SYS_REBOOT_NORMAL)

        else:
            return self._reboot_cmd(self._protocol.Reboot())

    def safe_mode(self):
        if self._client:
            return self.set_key('reboot', SYS_REBOOT_SAFE)

        else:
            return self._reboot_cmd(self._protocol.SafeMode())

    def reboot_and_load_fw(self):
        if self._client:
            return self.set_key('reboot', SYS_REBOOT_LOADFW)

        else:
            return self._reboot_cmd(self._protocol.LoadFirmware())

    def format_fs(self):
        if self._client:
            return self.set_key('reboot', SYS_REBOOT_FORMATFS)

        else:
            return self._send_command(self._protocol.FormatFS())

    def reboot_recovery(self):
        if self._client:
            return self.set_key('reboot', SYS_REBOOT_RECOVERY)

        else:
            raise Exception("Function not supported")

    def get_file_id(self, name):
        result = self._send_command(self._protocol.GetFileID(name=name))

        if result.file_id < 0:
            raise IOError("File: %s not found" % (name))

        return result.file_id

    def create_file(self, name):
        result = self._send_command(self._protocol.CreateFile(name=name))

        if result.file_id < 0:
            raise IOError("File: %s not created" % (name))

        return result.file_id

    def read_file_data(self, file_id, pos, length):
        result = self._send_command(self._protocol.ReadFileData(file_id=file_id, position=pos, length=length))

        return result.data

    def write_file_data(self, file_id, pos, data):
        result = self._send_command(self._protocol.WriteFileData(file_id=file_id, position=pos, length=len(data), data=data))

        return result.write_length

    def remove_file(self, file_id, timeout=None):
        result = self._send_command(self._protocol.RemoveFile(file_id=file_id), timeout=timeout)

        if result.status < 0:
            raise IOError("File: %s not deleted" % (file_id))

    def delete_file(self, filename):
        try:
            self._client.delete_file(filename)

        except (NoResponseFromHost, IOError, AttributeError, ProtocolErrorException):
            print "fallback to legacy", 2
            self._client = None
        
            file_id = self.get_file_id(filename)

            if filename in ['firmware.bin', 'firmware2.bin', 'error_log.txt']: # these files are slow to remove
                self.remove_file(file_id, timeout=5.0)

            else:
                self.remove_file(file_id)

    def get_file(self, filename, progress=None):
        try:
            data = self._client.read_file(filename, progress=progress)

        except (NoResponseFromHost, AttributeError, ProtocolErrorException):
            print "fallback to legacy", 3
            self._client = None
            file_id = self.get_file_id(filename)

            data = ""
            pos = 0

            while True:
                if progress:
                    progress(pos)

                result_data = self.read_file_data(file_id, pos, FILE_TRANSFER_LEN)

                data += result_data

                if len(result_data) < FILE_TRANSFER_LEN:
                    break

                pos += FILE_TRANSFER_LEN

        if progress:
            progress(len(data))

        return data

    def put_file(self, filename, data, progress=None):
        try:
            data = self._client.write_file(filename, data, progress=progress)

        except (NoResponseFromHost, AttributeError, ProtocolErrorException) as e:
            print "fallback to legacy", 4, e
            self._client = None
            # get file id
            try:
                file_id = self.get_file_id(filename)

                # remove file if it exists.
                # otherwise, we will overwrite the existing data, but if the old data
                # is longer than the new data, we'll still have the old data at the end
                # of the file.
                self.remove_file(file_id)
                
            except IOError:
                pass

            # create file
            file_id = self.create_file(filename)

            pos = 0

            while pos < len(data):
                chunk = data[pos:pos + FILE_TRANSFER_LEN]

                if progress:
                    progress(pos)

                if self.write_file_data(file_id, pos, chunk) < len(chunk):
                    raise IOError("Write error occurred :-(")

                pos += FILE_TRANSFER_LEN

        if progress:
            progress(len(data))

    def list_files_raw(self):
        data = self.get_file("fileinfo")

        fileinfo = sapphiredata.FileInfoArray()

        fileinfo.unpack(data)

        return fileinfo

    def list_files(self):
        try:
            return self._client.list_files()

        except (NoResponseFromHost):
            fileinfo = self.list_files_raw()

            d = {}

            # iterate over file listing, filtering out empty files
            for f in [f for f in fileinfo if f.filesize >= 0]:
                d[f.filename] = {'filename': f.filename,
                                 'flags': f.flags,
                                 'size': f.filesize}

            return d

    def check_file(self, filename, data):
        filehash = catbus_string_hash(data)

        try:
            if self._client.check_file(filename)['hash'] != filehash:
                raise IOError("Firmware image does not match!")
                return

        except (NoResponseFromHost, IOError, AttributeError):
            print "fallback to legacy", 5
            self._client = None
            # read back to verify
            read_back = self.get_file(filename)

            if data != read_back:
                raise IOError("Firmware image does not match!")
                return

    def load_firmware(self, firmware_id=None, progress=None, verify=True):
        if firmware_id == None:
            fw_info = self.get_firmware_info()
            fw_file = firmware.get_firmware(fw_info.firmware_id)

        else:
            fw_file = firmware.get_firmware(firmware_id)

        if fw_file is None:
            raise IOError("Firmware image not found")

        # read firmware data
        f = open(fw_file, 'rb')
        firmware_data = f.read()
        f.close()

        # delete old firmware
        try:
            self.delete_file("firmware.bin")

        except (NoResponseFromHost, IOError, AttributeError):
            print "fallback to legacy", 6
            self._client = None
            file_id = self.get_file_id("firmware.bin")
            self.remove_file(file_id, timeout=5.0) # firmware delete is slow

        filehash = catbus_string_hash(firmware_data)

        # load firmware image
        self.put_file("firmware.bin", firmware_data, progress=progress)

        if verify:
            self.check_file("firmware.bin", firmware_data)

        # reboot to loader
        self.reboot_and_load_fw()

    def get_firmware_info(self):
        data = self.get_file("fwinfo")

        # create hash of firmware info
        h = hashlib.new('sha256')
        h.update(data)
        self._firmware_info_hash = h.hexdigest()

        # unpack
        fw_info = sapphiredata.FirmwareInfoField()
        fw_info.unpack(data)

        # update state
        self.firmware_id        = fw_info.firmware_id
        self.firmware_name      = fw_info.firmware_name
        self.firmware_version   = fw_info.firmware_version
        self.os_name            = fw_info.os_name
        self.os_version         = fw_info.os_version

        return fw_info

    def get_gc_info(self):
        data = self.get_file("gc_data")

        gc_array = ArrayField(_field=Uint32Field).unpack(data)

        return {"sector_erase_counts": gc_array}

    def get_kvlink_info(self):
        data = self.get_file("kvlinks")

        info = sapphiredata.KVLinkArray()
        info.unpack(data)

        return info

    def get_kvsend_info(self):
        data = self.get_file("kvsend")

        info = sapphiredata.KVSendArray()
        info.unpack(data)

        return info

    def get_kvreceivecache_info(self):
        data = self.get_file("kvrxcache")

        info = sapphiredata.KVReceiveCacheArray()
        info.unpack(data)

        return info

    def get_thread_info(self):
        data = self.get_file("threadinfo")

        info = sapphiredata.ThreadInfoArray()
        info.unpack(data)

        return info

    def get_thread_info_dump(self):
        data = self.get_file("threadinfo.dump")

        info = sapphiredata.ThreadInfoArray()
        info.unpack(data)

        return info

    def get_handle_info(self):
        data = self.get_file("handleinfo")

        info = sapphiredata.HandleInfoArray()
        info.unpack(data)

        return info

    def get_dns_info(self):
        data = self.get_file("dns_cache")
        info = sapphiredata.DnsCacheArray()
        info.unpack(data)

        return info

    def get_event_log(self):
        data = self.get_file("event_log")
        info = sapphiredata.EventArray()
        info.unpack(data)

        return info


    ##########################
    # Command Line Interface
    ##########################
    def cli_scan(self, line):
        self.scan()

        return "Done"

    def cli_echo(self, line):
        start = time.time()

        self.echo(line)

        elapsed = time.time() - start

        return "(%d ms)" % (elapsed * 1000)

    def cli_reboot(self, line):
        self.reboot()
        return "Rebooting"

    def cli_safemode(self, line):
        self.safe_mode()
        return "Rebooting into safe mode"

    def cli_formatfs(self, line):
        self.format_fs()

        return "Formatting file system and rebooting..."

    def cli_rm(self, line):
        self.delete_file(line)

        return "Removed: %s" % (line)

    def cli_ls(self, line):
        if self._client:
            fileinfo = self.list_files()

            s = "\n"

            # sort between disk and virtual files
            disk_files = sorted([f for f in fileinfo.values() if (f['flags'] & 1) == 0], key=lambda a: a['filename'])
            vfiles = sorted([f for f in fileinfo.values() if (f['flags'] & 1) != 0], key=lambda a: a['filename'])

            # iterate over file listing
            for f in disk_files:
                v = ''

                if f['flags'] == 1:
                    v = 'V'

                s += "%1s %6d %s\n" % \
                    (v,
                     f['size'],
                     f['filename'])

            for f in vfiles:
                v = ''

                if f['flags'] == 1:
                    v = 'V'

                s += "%1s %6d %s\n" % \
                    (v,
                     f['size'],
                     f['filename'])

        else:
            fileinfo = self.list_files_raw()

            s = "\n"

            # iterate over file listing, filtering out empty files
            for f in [f for f in fileinfo if f.filesize >= 0]:
                v = ''

                if f.flags == 1:
                    v = 'V'

                s += "%1s %6d %s\n" % \
                    (v,
                     f.filesize,
                     f.filename)

        return s

    def cli_cat(self, line):
        data = self.get_file(line)

        s = "\n" + data

        return s

    def cli_getfile(self, line):
        def progress(length):
            sys.stdout.write("\rReading: %5d bytes" % (length))
            sys.stdout.flush()

        print ""

        data = self.get_file(line, progress=progress)

        f = open(line, 'w')
        f.write(data)
        f.close()

        return ""

    def cli_putfile(self, line):
        def progress(length):
            sys.stdout.write("\rWrite: %5d bytes" % (length))
            sys.stdout.flush()

        f = open(line, 'rb')
        data = f.read()
        f.close()

        print ""

        self.put_file(line, data, progress=progress)

        return ""

    def cli_loadfw(self, line):
        def progress(length):
            sys.stdout.write("\rWrite: %5d bytes" % (length))
            sys.stdout.flush()

        if line == "":
            fw = None
        else:
            fw = line

        self.load_firmware(firmware_id=fw, progress=progress)

        print ""

        return "Rebooting..."

    def cli_rebootloadfw(self, line):
        self.reboot_and_load_fw()

        print ""

        return "Rebooting with load firmware command..."

    def cli_fwinfo(self, line):
        fwinfo = self.get_firmware_info()

        s = "App:%24s Ver:%s OSVer:%s ID:%s Size:%d" % \
            (fwinfo.firmware_name,
             fwinfo.firmware_version,
             fwinfo.os_version,
             fwinfo.firmware_id,
             fwinfo.firmware_length)

        return s

    def cli_linkinfo(self, line):
        info = self.get_kvlink_info()

        s = "\nFlags Source     Dest       Query ->\n"

        KV_MSG_LINK_FLAG_SOURCE             = 0x01

        for n in info:
            flags = ''

            if n.flags & KV_MSG_LINK_FLAG_SOURCE:
                flags += 'S'

            else:
                flags += '-'

            flags += '-'
            flags += '-'
            flags += '-'

            s += "%5s %-10d %-10d %-10d %-10d %-10d %-10d %-10d %-10d %-10d %-10d" % \
                (flags,
                 n.source_hash,
                 n.dest_hash,
                 n.query[0],
                 n.query[1],
                 n.query[2],
                 n.query[3],
                 n.query[4],
                 n.query[5],
                 n.query[6],
                 n.query[7])

            s += "\n"

        return s

    def cli_sendinfo(self, line):
        info = self.get_kvsend_info()

        s = "\nTTL IP:             Port  Source       Dest         \n"

        for n in info:
            s += "%3d %15s %5u %-12d %-12d" % \
                (n.ttl,
                 n.ip,
                 n.port,
                 n.source_hash,
                 n.dest_hash)

            s += "\n"

        return s

    def cli_rxcacheinfo(self, line):
        info = self.get_kvreceivecache_info()

        s = "\nTTL IP:             Port  Dest         Data         Sequence\n"

        for n in info:
            s += "%3d %15s %5u %-12d %-12d %5u" % \
                (n.ttl,
                 n.ip,
                 n.port,
                 n.dest_hash,
                 n.data,
                 n.sequence)


            s += "\n"

        return s

    def cli_gcinfo(self, line):
        info = self.get_gc_info()

        i = 0
        sectors = []

        for sector in info["sector_erase_counts"]:
            sectors.append(sector)

        s = "Erases:%7d Least:%6d Most:%6d" % (sum(sectors), min(sectors), max(sectors))

        if line == "all":

            i = 0
            s += "\nSector erase counts:\n"

            for sector in sectors:
                i += 1

                s += "%6d " % (sector)

                if (i % 8) == 0:
                    s += "\n"

        return s

    def cli_threadinfo(self, line):
        if line == 'dump':
            info = self.get_thread_info_dump()

        else:
            info = self.get_thread_info()

        s = "\nAddr  Line  Flags  Data         Time     Runs  Avg      Alarm     Name\n"

        for n in info:

            flags = ''

            if n.flags & 1:
                flags += 'W'
            else:
                flags += '-'

            if n.flags & 2:
                flags += 'Y'
            else:
                flags += '-'

            if n.flags & 4:
                flags += 'S'
            else:
                flags += '-'

            if n.flags & 8:
                flags += 'I'
            else:
                flags += '-'

            if n.flags & 16:
                flags += 'A'
            else:
                flags += '-'

            try:
                avg_time = n.run_time / n.runs

            except ZeroDivisionError:
                avg_time = 0

            s += "%5x  %4d   %5s %4d %12d %8d %5d %12d %s\n" % \
                (n.addr,
                 n.line,
                 flags,
                 n.data_size,
                 n.run_time,
                 n.runs,
                 avg_time,
                 n.alarm,
                 n.name)

        return s

    def cli_handleinfo(self, line):
        info = self.get_handle_info()

        s = "\nType             Size\n"

        memtypes = {
            0: "unknown",
            1: "netmsg",
            5: "thread",
            6: "file",
            7: "socket",
            8: "socket_buffer",
            10: "log_buffer",
            11: "cmd_buffer",
            12: "cmd_reply_buffer",
            13: "file_handle",
            14: "fs_blocks",
            15: "kv_link",
            16: "kv_send",
            17: "kv_rx_cache",
        }

        total_size = 0
        handles = 0
        counts = {}

        for n in info:
            if n.handle_size == 0:
                continue

            if n.handle_type in memtypes:
                type_str = memtypes[n.handle_type]

            else:
                type_str = "unknown"

            if type_str not in counts:
                counts[type_str] = [0, 0]

            counts[type_str][0] += 1
            counts[type_str][1] += n.handle_size

            s += "%16s %d\n" % (type_str, n.handle_size)

            handles += 1
            total_size += n.handle_size

        s += "\n"

        for n in counts:
            s += "%16s %4d %4d\n" % (n, counts[n][0], counts[n][1])

        s += "\n"
        s += "Handles: %d\n" % (handles)
        s += "Total size: %d\n" % (total_size)

        return s

    def cli_dnsinfo(self, line):
        dnsinfo = self.get_dns_info()

        s = "\n"

        # iterate over DNS cache entries, filtering out the empty ones
        for d in [d for d in dnsinfo if d.status != 0]:
            if d.status == 1:
                status = "valid"
            else:
                status = "invalid"

            s += "IP:%15s Status:%8s TTL:%8d Query:%s" % \
                (d.ip,
                 status,
                 d.ttl,
                 d.query)

        return s

    def cli_eventlog(self, line):
        events = sorted(self.get_event_log(), key=lambda evt: evt.timestamp)

        if len(line) == 0:
            line = '*'

        eventlog = [event for event in events if fnmatch.fnmatch(event.event_str, line)]

        s = "\n"

        last_timestamp = 0

        # iterate over events
        for evt in eventlog:
            if evt.event_str in ['timer_alarm_miss',
                                 'cpu_wake',
                                 'func_enter',
                                 'func_exit',
                                 'thread_id']:

                if evt.event_str in ['thread_id', 'func_enter', 'func_exit']:
                    evt.param *= 2 # convert to byte address

                param_str = "0x%5x" % (evt.param)

            else:
                if evt.event_str in ['stack_pointer']:
                    evt.param -= 512
                    evt.param = 16384 - evt.param

                param_str = "%7d" % (evt.param)


            s += "%32s: %s    @ %16s +%16s\n" % \
                (evt.event_str,
                 param_str,
                 timedelta(microseconds=evt.timestamp),
                 timedelta(microseconds=(evt.timestamp - last_timestamp)))

            last_timestamp = evt.timestamp

        return s

    def cli_clearlogs(self, line):
        for filename in ["log.txt", "threadinfo.dump", "event_log", "error_log.txt"]:
            try:
                self.delete_file(filename)

            except IOError:
                pass


    def cli_getkey(self, line):
        if line == "":
            params = self.get_kv('*')

        else:
            params = self.get_kv(line)

        if isinstance(params, dict):
            s = "\nName                             Flags  Type     Value\n"

            for k in sorted(params.iterkeys()):
                s += "%s\n" % (self._keys[k])

        else:
            s = "%s = %s" % (line, params)

        return s

    def cli_setkey(self, line):
        tokens = line.split(' ', 1)
        param = tokens[0]
        value = tokens[1]
        
        # check for special blank value (for resetting strings to empty)
        if str(value) == '__null__':
            value = ''

        self.set_key(param, value)

        new_param = self.get_key(param)

        return "%s set to: %s" % (param, new_param)


    def cli_resetcfg(self, line):
        print ""
        print "DANGER ZONE! Are you sure you want to do this?"
        print "Type 'yes' if you are sure."

        yes = raw_input()

        if yes == "yes":
            self.reset_config()

            return "Configuration reset"

        return "No changes made"

    def cli_systime(self, line):
        t = self.get_key("sys_time")
        dt = timedelta(seconds=long(t / 1000))

        return "%11d ms (%s)" % (t, str(dt))

    def cli_status(self, line):
        params = self.get_kv("sys_mode", "mem_peak_usage", "fs_free_space", "sys_time", "supply_voltage")

        if params["sys_mode"] == 0:
            mode = "normal "

        elif params["sys_mode"] == 1:
            mode = "safe   "

        else:
            mode = "unknown"

        s = "Mode:%s PeakMem:%4d DiskSpace:%6d SysTime:%11d Volts:%3.2f" % (mode,
                                                                            params["mem_peak_usage"],
                                                                            params["fs_free_space"],
                                                                            params["sys_time"],
                                                                            params["supply_voltage"] / 1000.0)

        return s

    def cli_warnings(self, line):
        params = self.get_key("sys_warnings")

        warnings = decode_warnings(params)

        s = ''
        for w in warnings:
            s += w + " "

        if len(warnings) == 0:
            s = "OK"

        return s

    def cli_diskinfo(self, line):
        params = self.get_kv("fs_free_space", "fs_total_space", "fs_disk_files", "fs_max_disk_files", "fs_virtual_files", "fs_max_virtual_files")

        s = "Free:%6d Total:%6d Files:%2d / %2d Vfiles:%2d / %2d" % \
            (params["fs_free_space"],
             params["fs_total_space"],
             params["fs_disk_files"],
             params["fs_max_disk_files"],
             params["fs_virtual_files"],
             params["fs_max_virtual_files"])

        return s

    def cli_cpuinfo(self, line):
        params = self.get_kv("thread_task_time", "thread_sleep_time", "thread_peak", "thread_loops", "thread_run_time")

        # convert all params to floats
        params = {k: float(v) for (k, v) in params.iteritems()}

        if params["thread_run_time"] == 0:
            loop_rate = 0
            cpu_usage = 0

        else:
            loop_rate = params["thread_loops"] / (params["thread_run_time"] / 1000.0)
            cpu_usage = (params["thread_task_time"] / params["thread_run_time"]) * 100.0

        s = "CPU:%2.1f%% Tsk:%8d Slp:%8d Ohd:%8d MaxT:%3d Loops:%8d @ %5d/sec" % \
            (cpu_usage,
             params["thread_task_time"],
             params["thread_sleep_time"],
             params["thread_run_time"] - (params["thread_task_time"] + params["thread_sleep_time"]),
             params["thread_peak"],
             params["thread_loops"],
             loop_rate)

        return s

    def cli_irqinfo(self, line):
        params = self.get_kv("sys_time_us", "irq_time", "irq_longest_time", "irq_longest_addr")

        # convert all params to floats
        params = {k: float(v) for (k, v) in params.iteritems()}

        irq_usage = (params["irq_time"] / params["sys_time_us"]) * 100.0

        s = "IRQ:%2.1f%% Longest:%6d uS Addr:0x%05x" % \
            (irq_usage,
             params["irq_longest_time"],
             params["irq_longest_addr"])

        return s


    def cli_meminfo(self, line):
        params = self.get_kv("mem_handles", "mem_max_handles", "mem_stack", "mem_max_stack", "mem_free_space", "mem_peak_usage", "mem_heap_size", "mem_used", "mem_dirty")

        s = "Handles:%3d/%3d Stack:%4d/%4d Free:%4d Used:%4d Dirty: %4d Peak:%4d Heap:%4d" % \
             (params["mem_handles"],
              params["mem_max_handles"],
              params["mem_stack"],
              params["mem_max_stack"],
              params["mem_free_space"],
              params["mem_used"],
              params["mem_dirty"],
              params["mem_peak_usage"],
              params["mem_heap_size"])

        return s

    def cli_loaderinfo(self, line):
        params = self.get_kv("loader_status", "loader_version_minor", "loader_version_major")

        s = "Loader version: %s.%s Status: %d" % \
            (chr(params["loader_version_major"]),
             chr(params["loader_version_minor"]),
             params["loader_status"])

        return s

    def cli_ntptime(self, line):
        ntp_seconds = self.get_key("ntp_seconds")

        ntp_now = timedelta(seconds=ntp_seconds) + NTP_EPOCH

        return ntp_now.isoformat()

    def cli_loadwifi(self, line):
        def progress(length):
            sys.stdout.write("\rWrite: %5d bytes" % (length))
            sys.stdout.flush()

        filename = 'wifi_firmware.bin'
        with open(filename, 'rb') as f:
            data = f.read()

        data_bytes = [ord(c) for c in data]

        # verify first byte (quick sanity check)
        if data_bytes[0] != 0xE9:
            print "Invalid firmware image!"
            return

        # compute firmware length, minus the md5 at the end
        # md5 is always 16 bytes
        fw_len = len(data) - 16

        # get md5 from file
        file_md5 = data[fw_len:]
        md5_digest = hashlib.md5(data[:fw_len]).digest()

        if file_md5 != md5_digest:
            print "Invalid firmware image!"
            return



        # preprocessing should already be done by build script,
        # but leaving this here for now in case we need it.

        # # we override bytes 2 and 3 in the ESP8266 image
        # data_bytes[2] = 0
        # data_bytes[3] = 0
        #
        # # convert back to string
        # data = ''.join(map(chr, data_bytes))
        #
        # # need to pad to sector length
        # padding_len = 4096 - (len(data) % 4096)
        # data += (chr(0xff) * padding_len)
        #
        # fw_len = len(data)
        #
        # md5_digest = hashlib.md5(data).digest()
        # data += md5_digest
        #
        # with open('wifi_firmware_padded.bin', 'w') as f:
        #     f.write(data)



        print "\nLoading firmware image"

        try:
            self.delete_file(filename)

        except IOError: # file not found
            pass

        time.sleep(2.0) # give a second while file system erases blocks

        # calculate crc of file data
        filehash = catbus_string_hash(data)

        try:
            self.put_file(filename, data, progress=progress)

        except Exception as e:
            print type(e), e
            raise

        print "\nVerifying firmware image..."

        self.check_file(filename, data)

        print "Setting firmware length..."
        self.set_key('wifi_fw_len', fw_len)

        print "Setting MD5..."
        self.set_key('wifi_md5', binascii.hexlify(md5_digest))

        print "Starting wifi firmware flasher..."

        self.reboot()

        time.sleep(5.0)

        for i in xrange(20):
            try:
                self.echo('')

            except DeviceUnreachableException:
                time.sleep(1.0)


        self.echo('')

        print "Firmware load complete"

    def cli_loadvm(self, line):
        with open('vm.bin', 'rb') as f:
            data = f.read()

        try:
            self.delete_file('vm.bin')

        except IOError:
            pass

        self.put_file('vm.bin', data)

        self.set_key('vm_run', True)
        self.set_key('vm_reset', True)

    def cli_dumpkvmeta(self, line):
        data = self.get_file("kvmeta")
        kvmeta = sapphiredata.KVMetaArray().unpack(data)

        for kv in kvmeta:
            print kv

    def cli_getfileid(self, line):
        file_id = self.get_file_id(line)

        print "File ID: %d" % (file_id)

def createDevice(**kwargs):
    return Device(**kwargs)



# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

from datetime import datetime, timedelta
import logging
import logging.handlers
import socket
import os


EPOCH = datetime.utcfromtimestamp(0)
NTP_EPOCH = datetime(1900, 1, 1)


def memoize(f):
    """ Memoization decorator for a function taking a single argument """
    class memodict(dict):
        def __missing__(self, key):
            ret = self[key] = f(key)
            return ret
    return memodict().__getitem__



def iso_to_datetime(iso):
    tries = ["%Y-%m-%dT%H:%M:%S.%f",
             "%Y-%m-%dT%H:%M:%S",
             "%Y-%m-%dT%H:%M:%S.%fZ",
             "%Y-%m-%dT%H:%M:%SZ",
             "%Y-%m-%d %H:%M:%S.%f",
             "%Y-%m-%d %H:%M:%S",
             "%Y-%m-%d %H:%M:%S.%fZ",
             "%Y-%m-%d %H:%M:%SZ"]

    for trie in tries:
        try:
            return datetime.strptime(iso, trie)

        except ValueError:
            pass

    raise ValueError

def coerce_to_list(obj):
    if not hasattr(obj, '__iter__'):
        return [obj]

    return obj

def unix_time(dt):
    delta = dt - EPOCH
    return delta.total_seconds()


def unix_time_millis(dt):
    return unix_time(dt) * 1000.0

def ts_to_datetime(ts):
    try:
        return datetime.utcfromtimestamp(ts)

    except TypeError:
        # already a datetime object
        return ts

def convert_string_to_more_specific_type(value):
    # try to convert to a more specific type
    if value.lower() == 'true':
        return True

    elif value.lower() == 'false':
        return False

    else:
        try:
            return int(value)

        except ValueError:
            try:
                return float(value)

            except ValueError:
                # leave as string
                return value

def coerce_value_to_target_type(value, target):
    if isinstance(target, type(value)):
        return value

    elif isinstance(target, bool):
        return bool(value)

    elif isinstance(target, int):
        return int(value)

    elif isinstance(target, float):
        return float(value)

    elif isinstance(target, long):
        return long(value)

    elif isinstance(target, basestring):
        return str(value)

    return value

def now():
    return datetime.utcnow()

def datetime_to_ntp(dt):
    time_diff = dt - NTP_EPOCH
    seconds = int(time_diff.total_seconds())
    fraction = int((time_diff.total_seconds() - seconds) * (pow(2, 32) - 1))

    return seconds, fraction

def ntp_to_datetime(seconds, fraction):
    fraction = float(fraction) / (pow(2, 32) - 1)
    seconds = float(seconds) + fraction

    delta = timedelta(seconds=seconds)

    return delta + NTP_EPOCH

def datetime_to_microseconds(dt):
    return long((dt - NTP_EPOCH).total_seconds() * 1000000)

def microseconds_to_datetime(ms):
    return timedelta(microseconds=ms) + NTP_EPOCH

def hostname():
    socket.gethostname()

def ip():
    return socket.gethostbyname(socket.gethostname())

def linear_interp(x, x0, x1, y0, y1):
    if x > x1:
        x = x1

    elif x < x0:
        x = x0

    y = y0 + (y1 - y0)*((x - x0) / (x1 - x0))

    return y

def coerce_string_to_list(s):
    if isinstance(s, basestring):
        return [s]

    return s

def is_wildcard(key):
        if '*' in key or \
           '?' in key or \
           '[' in key:
           return True

        return False

logging_initalized = False

def setup_basic_logging(console=True, filename=None):
    # dt_format = '%Y-%m-%dT%H:%M:%S'
    # logging.basicConfig(format='%(asctime)s >>> %(levelname)s (%(threadName)s) %(message)s',
    #                     datefmt=dt_format,
    #                     level=logging.DEBUG)

    global logging_initalized

    if logging_initalized:
        return

    logging_initalized = True


    dt_format = '%Y-%m-%dT%H:%M:%S'

    root = logging.getLogger('')
    root.setLevel(logging.DEBUG)

    if console:
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s %(asctime)s.%(msecs)03d %(threadName)s %(message)s', datefmt=dt_format)
        handler.setFormatter(formatter)
        root.addHandler(handler)

    if filename:
        # get path
        path, name = os.path.split(filename)

        if len(path) > 0:
            # if directory does not exist
            if not os.path.exists(path):
                os.makedirs(path)

        handler = logging.handlers.RotatingFileHandler(filename, maxBytes=32*1048576, backupCount=4)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s %(asctime)s.%(msecs)03d %(threadName)s %(message)s', datefmt=dt_format)
        handler.setFormatter(formatter)
        root.addHandler(handler)


    # create a network logger
    # socket_handler = logging.handlers.SocketHandler('localhost',
    #                     logging.handlers.DEFAULT_TCP_LOGGING_PORT)

    # root.addHandler(socket_handler)
